# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import pprint as pp
import re
import six
import xml.etree.ElementTree as ET

from oslo_config import cfg
from oslo_log import log as logging

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k import (
    asr1k_snippets as asr_snippets)
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.common.htparser import HTParser
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerrole

LOG = logging.getLogger(__name__)


ROUTER_ROLE_ATTR = routerrole.ROUTER_ROLE_ATTR
ROUTER_ROLE_HA_REDUNDANCY = cisco_constants.ROUTER_ROLE_HA_REDUNDANCY


NROUTER_REGEX = "nrouter-(\w{6,6})"
NROUTER_MULTI_REGION_REGEX = "nrouter-(\w{6,6})-(\w{1,7})"

VRF_REGEX = "ip vrf " + NROUTER_REGEX
VRF_REGEX_NEW = "vrf definition " + NROUTER_REGEX

VRF_MULTI_REGION_REGEX = "ip vrf " + NROUTER_MULTI_REGION_REGEX
VRF_MULTI_REGION_REGEX_NEW = "vrf definition " + NROUTER_MULTI_REGION_REGEX

INTF_REGEX = "interface \S+\.(\d+)"
INTF_DESC_REGEX = (
    "\s*description OPENSTACK_NEUTRON_INTF|OPENSTACK_NEUTRON_EXTERNAL_INTF")
INTF_DESC_MULTI_REGION_REGEX = (
    "\s*description OPENSTACK_NEUTRON_(\w{1,7})_INTF|"
    "OPENSTACK_NEUTRON_EXTERNAL_(\w{1,7})_INTF")
VRF_EXT_INTF_REGEX = "\s*ip vrf forwarding .*"
VRF_INTF_REGEX = "\s*ip vrf forwarding " + NROUTER_REGEX
VRF_INTF_MULTI_REGION_REGEX = ("\s*ip vrf forwarding " +
    NROUTER_MULTI_REGION_REGEX)
VRF_EXT_INTF_REGEX_NEW = "\s*vrf forwarding .*"
VRF_INTF_REGEX_NEW = "\s*vrf forwarding " + NROUTER_REGEX
VRF_INTF_MULTI_REGION_REGEX_NEW = ("\s*vrf forwarding " +
    NROUTER_MULTI_REGION_REGEX)
DOT1Q_REGEX = "\s*encapsulation dot1Q (\d+)"
INTF_NAT_REGEX = "\s*ip nat (inside|outside)"
HSRP_REGEX = "\s*standby (\d+) .*"

INTF_V4_ADDR_REGEX = ("\s*ip address (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})")
HSRP_V4_VIP_REGEX = "\s*standby (\d+) ip (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"

SNAT_REGEX = ("ip nat inside source static"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) vrf " +
    NROUTER_REGEX +
    " redundancy neutron-hsrp-(\d+)-(\d+)")

SNAT_MULTI_REGION_REGEX = ("ip nat inside source static"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " vrf " + NROUTER_MULTI_REGION_REGEX +
    " redundancy neutron-hsrp-(\d+)-(\d+)")

SNAT_REGEX_OLD = ("ip nat inside source static"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) vrf " +
    NROUTER_REGEX +
    " redundancy neutron-hsrp-grp-(\d+)-(\d+)")

SNAT_MULTI_REGION_REGEX_OLD = ("ip nat inside source static"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) vrf " +
    NROUTER_MULTI_REGION_REGEX +
    " redundancy neutron-hsrp-grp-(\d+)-(\d+)")

NAT_POOL_REGEX = ("ip nat pool " +
    NROUTER_REGEX +
    "_nat_pool (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) netmask"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})")
NAT_POOL_MULTI_REGION_REGEX = ("ip nat pool " + NROUTER_MULTI_REGION_REGEX +
    "_nat_pool (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) netmask"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})")

NAT_OVERLOAD_REGEX = ("ip nat inside source list neutron_acl_(\d+)"
    " interface \S+\.(\d+) vrf " +
    NROUTER_REGEX +
    " overload")
NAT_OVERLOAD_MULTI_REGION_REGEX = ("ip nat inside source list"
    " neutron_acl_(\d+) interface \S+\.(\d+) vrf " +
    NROUTER_MULTI_REGION_REGEX + " overload")

NAT_POOL_OVERLOAD_REGEX = ("ip nat inside source list"
    " neutron_acl_(\d+)_(\w{1,8}) pool " +
    NROUTER_REGEX +
    "_nat_pool vrf " +
    NROUTER_REGEX +
    " overload")
NAT_POOL_OVERLOAD_MULTI_REGION_REGEX = ("ip nat inside source"
    " list neutron_acl_(\w{1,7})_(\d+)_(\w{1,8}) pool " +
    NROUTER_MULTI_REGION_REGEX +
    "_nat_pool vrf " +
    NROUTER_MULTI_REGION_REGEX +
    " overload")

ACL_REGEX = "ip access-list standard neutron_acl_(\d+)_(\w{1,8})"
ACL_MULTI_REGION_REGEX = ("ip access-list standard neutron_acl_" +
                          "(\w{1,7})_(\d+)_(\w{1,8})")
ACL_CHILD_REGEX = ("\s*permit (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})")

DEFAULT_ROUTE_REGEX = ("ip route vrf " +
    NROUTER_REGEX + " 0\.0\.0\.0 0\.0\.0\.0 \S+\.(\d+)"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})")
DEFAULT_ROUTE_MULTI_REGION_REGEX = ("ip route vrf " +
    NROUTER_MULTI_REGION_REGEX +
    " 0\.0\.0\.0 0\.0\.0\.0 \S+\.(\d+)"
    " (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})")

DEFAULT_ROUTE_V6_REGEX_BASE = ("ipv6 route vrf " +
    NROUTER_REGEX +
    " ::/0 %s\.(\d+) ([0-9A-Fa-f:]+) nexthop-vrf default")
DEFAULT_ROUTE_V6_MULTI_REGION_REGEX_BASE = ("ipv6 route vrf " +
    NROUTER_MULTI_REGION_REGEX +
    " ::/0 %s\.(\d+) ([0-9A-Fa-f:]+) nexthop-vrf default")

TENANT_ROUTE_V6_REGEX_BASE = ("ipv6 route ([0-9A-Fa-f:/]+)"
    " %s.(\d+) nexthop-vrf " +
    NROUTER_REGEX)
TENANT_ROUTE_V6_MULTI_REGION_REGEX_BASE = ("ipv6 route ([0-9A-Fa-f:/]+)"
    " %s.(\d+) nexthop-vrf " +
    NROUTER_MULTI_REGION_REGEX)

INTF_V6_ADDR_REGEX = "\s*ipv6 address ([0-9A-Fa-f:]+)\/(\d+)"


XML_FREEFORM_SNIPPET = "<config><cli-config-data>%s</cli-config-data></config>"
XML_CMD_TAG = "<cmd>%s</cmd>"


def is_port_v6(port):
    prefix = port['subnets'][0]['cidr']
    if netaddr.IPNetwork(prefix).version == 6:
        return True
    else:
        return False


class ConfigSyncer(object):

    def __init__(self, router_db_info, driver,
                 hosting_device_info, test_mode=False):
        self.existing_cfg_dict = {}
        self.driver = driver
        self.hosting_device_info = hosting_device_info
        self.existing_cfg_dict['interfaces'] = {}
        self.existing_cfg_dict['dyn_nat'] = {}
        self.existing_cfg_dict['static_nat'] = {}
        self.existing_cfg_dict['acls'] = {}
        self.existing_cfg_dict['routes'] = {}
        self.existing_cfg_dict['pools'] = {}

        self.segment_gw_dict = {}

        router_id_dict, interface_segment_dict, segment_nat_dict = \
            self.process_routers_data(router_db_info)
        self.router_id_dict = router_id_dict
        self.intf_segment_dict = interface_segment_dict
        self.segment_nat_dict = segment_nat_dict
        self.test_mode = test_mode
        if (cfg.CONF.multi_region.enable_multi_region):
            self.route_regex = DEFAULT_ROUTE_MULTI_REGION_REGEX
        else:
            self.route_regex = DEFAULT_ROUTE_REGEX

    def process_routers_data(self, routers):
        hd_id = self.hosting_device_info['id']
        router_id_dict = {}
        interface_segment_dict = {}
        segment_nat_dict = {}

        for router in routers:
            if 'hosting_device' not in router:
                continue

            if router['hosting_device']['id'] != hd_id:
                continue

            short_router_id = router['id'][0:6]
            router_id_dict[short_router_id] = router

            interfaces = []
            if '_interfaces' in router.keys():
                interfaces += router['_interfaces']

            # Orgnize interfaces, indexed by segment_id
            for interface in interfaces:
                hosting_info = interface['hosting_info']
                segment_id = hosting_info['segmentation_id']
                if segment_id not in interface_segment_dict:
                    interface_segment_dict[segment_id] = []
                    if segment_id not in segment_nat_dict:
                        segment_nat_dict[segment_id] = False
                if (router[ROUTER_ROLE_ATTR] ==
                        cisco_constants.ROUTER_ROLE_GLOBAL):
                    interface['is_external'] = True
                    if (segment_id not in self.segment_gw_dict):
                        self.segment_gw_dict[segment_id] = [interface]
                else:
                    interface['is_external'] = False
                interface_segment_dict[segment_id].append(interface)

            # Mark which segments have NAT enabled
            # i.e., the segment is present on at least 1 router with
            # both external and internal networks present
            if 'gw_port' in router.keys():
                gw_port = router['gw_port']
                gw_segment_id = gw_port['hosting_info']['segmentation_id']

                if (router[ROUTER_ROLE_ATTR] ==
                    cisco_constants.ROUTER_ROLE_GLOBAL):

                    if (gw_segment_id not in self.segment_gw_dict):
                        self.segment_gw_dict[gw_segment_id] = [gw_port]

                if '_interfaces' in router.keys():
                    interfaces = router['_interfaces']
                    for intf in interfaces:
                        if (intf['device_owner'] ==
                                bc.constants.DEVICE_OWNER_ROUTER_INTF):
                            if is_port_v6(intf) is not True:
                                intf_segment_id = (
                                    intf['hosting_info']['segmentation_id'])
                                segment_nat_dict[gw_segment_id] = True
                                segment_nat_dict[intf_segment_id] = True

        return router_id_dict, interface_segment_dict, segment_nat_dict

    def delete_invalid_cfg(self, conn=None):
        router_id_dict = self.router_id_dict
        intf_segment_dict = self.intf_segment_dict
        segment_nat_dict = self.segment_nat_dict

        if not conn:
            conn = self.driver._get_connection()

        LOG.info("neutron router db records")

        for router_id, router in six.iteritems(router_id_dict):
            LOG.info("ROUTER_ID: %s", router_id)

        LOG.info("\n")

        for segment_id, intf_list in six.iteritems(intf_segment_dict):
            LOG.info("SEGMENT_ID: %s", (segment_id))
            for intf in intf_list:
                dev_owner = intf['device_owner']
                dev_id = intf['device_id'][0:6]
                ip_addr = intf['fixed_ips'][0]['ip_address']
                intf_info = "    INTF: %s, %s, %s" % \
                            (ip_addr, dev_id, dev_owner)
                LOG.info(intf_info)

        running_cfg = self.get_running_config(conn)
        parsed_cfg = HTParser(running_cfg)

        invalid_cfg = []

        invalid_cfg += self.clean_snat(conn,
                                       router_id_dict,
                                       intf_segment_dict,
                                       segment_nat_dict,
                                       parsed_cfg)

        invalid_cfg += self.clean_nat_pool_overload(conn,
                                                    router_id_dict,
                                                    intf_segment_dict,
                                                    segment_nat_dict,
                                                    parsed_cfg)

        invalid_cfg += self.clean_nat_pool(conn,
                                           router_id_dict,
                                           intf_segment_dict,
                                           segment_nat_dict,
                                           parsed_cfg)

        invalid_cfg += self.clean_routes(conn,
                                         router_id_dict,
                                         intf_segment_dict,
                                         segment_nat_dict,
                                         parsed_cfg,
                                         self.route_regex)
        invalid_cfg += self.clean_acls(conn,
                                       intf_segment_dict,
                                       segment_nat_dict,
                                       parsed_cfg)

        invalid_cfg += self.clean_interfaces(conn,
                                             intf_segment_dict,
                                             segment_nat_dict,
                                             parsed_cfg)

        invalid_cfg += self.clean_vrfs(conn, router_id_dict, parsed_cfg)
        LOG.debug("invalid_cfg = %s " % (pp.pformat(invalid_cfg)))
        return invalid_cfg

    def get_running_config(self, conn):
        """Get the ASR1k's current running config.
        :return: Current IOS running config as multiline string
        """
        config = conn.get_config(source="running")
        if config:
            root = ET.fromstring(config._raw)
            running_config = root[0][0]
            rgx = re.compile("\r*\n+")
            ioscfg = rgx.split(running_config.text)
            return ioscfg

    def get_ostk_router_ids(self, router_id_dict):
        ostk_router_ids = []
        for router_id, router in six.iteritems(router_id_dict):
            ostk_router_ids.append(router_id)
        return ostk_router_ids

    def get_running_config_router_ids(self, parsed_cfg):
        rconf_ids = []
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region

        if (is_multi_region_enabled):
            vrf_regex_new = VRF_MULTI_REGION_REGEX_NEW
        else:
            vrf_regex_new = VRF_REGEX_NEW

        for parsed_obj in parsed_cfg.find_objects(vrf_regex_new):
            LOG.info("VRF object: %s", (str(parsed_obj)))
            match_obj = re.match(vrf_regex_new, parsed_obj.text)
            router_id = match_obj.group(1)
            LOG.info("    First 6 digits of router ID: %s\n",
                     (router_id))
            if (is_multi_region_enabled):
                region_id = match_obj.group(2)
                LOG.info("    region ID: %s\n",
                         (region_id))
                my_region_id = cfg.CONF.multi_region.region_id
                if (my_region_id == region_id):
                    rconf_ids.append(router_id)
            else:
                rconf_ids.append(router_id)

        return rconf_ids

    def clean_vrfs(self, conn, router_id_dict, parsed_cfg):
        ostk_router_ids = self.get_ostk_router_ids(router_id_dict)
        rconf_ids = self.get_running_config_router_ids(parsed_cfg)

        source_set = set(ostk_router_ids)
        dest_set = set(rconf_ids)

        # add_set = source_set.difference(dest_set)
        del_set = dest_set.difference(source_set)

        LOG.info("VRF DB set: %s", (source_set))
        LOG.info("VRFs to delete: %s", (del_set))
        # LOG.info("VRFs to add: %s" % (add_set))

        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        invalid_vrfs = []
        for router_id in del_set:
            if (is_multi_region_enabled):
                my_region_id = cfg.CONF.multi_region.region_id
                invalid_vrfs.append("nrouter-%s-%s" % (router_id,
                                                       my_region_id))
            else:
                invalid_vrfs.append("nrouter-%s" % (router_id))

        if not self.test_mode:
            for vrf_name in invalid_vrfs:
                confstr = asr_snippets.REMOVE_VRF_DEFN % vrf_name
                conn.edit_config(target='running', config=confstr)

        LOG.debug("invalid_vrfs = %s" % (pp.pformat(invalid_vrfs)))
        return invalid_vrfs

    def get_single_cfg(self, cfg_line):
        if len(cfg_line) != 1:
            return None
        else:
            return cfg_line[0]

    def clean_nat_pool(self,
                       conn,
                       router_id_dict,
                       intf_segment_dict,
                       segment_nat_dict,
                       parsed_cfg):
        delete_pool_list = []

        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        if (is_multi_region_enabled):
            nat_pool_regex = NAT_POOL_MULTI_REGION_REGEX
        else:
            nat_pool_regex = NAT_POOL_REGEX

        pools = parsed_cfg.find_objects(nat_pool_regex)
        for pool in pools:
            LOG.info("\nNAT pool: %s", (pool))
            match_obj = re.match(nat_pool_regex, pool.text)
            if (is_multi_region_enabled):
                router_id, region_id, start_ip, end_ip, netmask = (
                    match_obj.group(1, 2, 3, 4, 5))
                my_region_id = cfg.CONF.multi_region.region_id
                other_region_ids = cfg.CONF.multi_region.other_region_ids
                if region_id != my_region_id:
                    if region_id not in other_region_ids:
                        delete_pool_list.append(pool.text)
                    else:
                        # skip because some other deployment owns
                        # this configuration
                        continue
            else:
                router_id, start_ip, end_ip, netmask = (
                    match_obj.group(1, 2, 3, 4))

            # Check that VRF exists in openstack DB info
            if router_id not in router_id_dict:
                LOG.info("router not found for NAT pool, deleting")
                delete_pool_list.append(pool.text)
                continue

            # Check that router has external network
            router = router_id_dict[router_id]
            if "gw_port" not in router:
                LOG.info("router has no gw_port, pool is invalid,"
                         " deleting")
                delete_pool_list.append(pool.text)
                continue

            # Check IPs and netmask
            if router.get(ROUTER_ROLE_ATTR) == ROUTER_ROLE_HA_REDUNDANCY:
                the_port = router['gw_port'][ha.HA_INFO]['ha_port']
            else:
                the_port = router['gw_port']
            found = False
            for i in range(len(the_port['fixed_ips'])):
                pool_ip = str(the_port['fixed_ips'][i]['ip_address'])
                if start_ip == pool_ip:
                    found = True
                    break

            if found is False:
                LOG.info("start IP %(start_ip)s for "
                         "pool does not match %(pool_ip)s, deleting" %
                         {'start_ip': start_ip, 'pool_ip': pool_ip})
                delete_pool_list.append(pool.text)
                continue

            if end_ip != pool_ip:
                LOG.info("end IP for pool does not match, deleting")
                delete_pool_list.append(pool.text)
                continue

            pool_net = netaddr.IPNetwork(the_port['subnets'][i]['cidr'])
            if netmask != str(pool_net.netmask):
                LOG.info(
                    "netmask for pool does not match, netmask:%(netmask)s,"
                    " pool_netmask:%(pool_netmask)s, deleting" %
                    {'netmask': netmask, 'pool_netmask': pool_net.netmask})
                delete_pool_list.append(pool.text)
                continue

            self.existing_cfg_dict['pools'][pool_ip] = pool

        if not self.test_mode:
            for pool_cfg in delete_pool_list:
                del_cmd = XML_CMD_TAG % ("no %s" % (pool_cfg))
                confstr = XML_FREEFORM_SNIPPET % (del_cmd)
                LOG.info("Delete pool: %s", del_cmd)
                conn.edit_config(target='running', config=confstr)
        LOG.debug("delete_pool_list = %s " % (pp.pformat(delete_pool_list)))
        return delete_pool_list

    def clean_routes(self,
                     conn,
                     router_id_dict,
                     intf_segment_dict,
                     segment_nat_dict,
                     parsed_cfg,
                     route_regex):
        delete_route_list = []
        routes = parsed_cfg.find_objects(route_regex)
        for route in routes:
            LOG.info("\ndefault route: %s", (route))
            match_obj = re.match(route_regex, route.text)
            is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
            if (is_multi_region_enabled):
                router_id, region_id, segment_id, next_hop = (
                    match_obj.group(1, 2, 3, 4))
                LOG.info("    router_id: %(router_id)s,"
                         ", region_id: %(region_id)s, segment_id:"
                         " %(segment_id)s, next_hop: %(next_hop)s" %
                         {'router_id': router_id,
                          'region_id': region_id,
                          'segment_id': segment_id,
                          'next_hop': next_hop})
                my_region_id = cfg.CONF.multi_region.region_id
                other_region_ids = cfg.CONF.multi_region.other_region_ids
                if region_id != my_region_id:
                    if region_id not in other_region_ids:
                        delete_route_list.append(route.text)
                    else:
                        # skip because some other deployment owns
                        # this configuration
                        continue
            else:
                router_id, segment_id, next_hop = (
                    match_obj.group(1, 2, 3))

            LOG.info("    router_id: %(router_id)s, segment_id:"
                     " %(segment_id)s, next_hop: %(next_hop)s" %
                     {'router_id': router_id,
                      'segment_id': segment_id,
                      'next_hop': next_hop})

            # Check that VRF exists in openstack DB info
            if router_id not in router_id_dict:
                LOG.info("router not found for route, deleting")
                delete_route_list.append(route.text)
                continue

            # Check that router has external network and segment_id matches
            router = router_id_dict[router_id]
            if "gw_port" not in router:
                LOG.info("router has no gw_port, route is invalid,"
                         " deleting")
                delete_route_list.append(route.text)
                continue

            gw_port = router['gw_port']
            gw_segment_id = gw_port['hosting_info']['segmentation_id']
            if int(segment_id) != gw_segment_id:
                LOG.info("route segment_id does not match router's gw"
                         " segment_id, deleting")
                delete_route_list.append(route.text)
                continue

            # Check that nexthop matches gw_ip of external network
            if router.get(ROUTER_ROLE_ATTR) == ROUTER_ROLE_HA_REDUNDANCY:
                the_port = router['gw_port'][ha.HA_INFO]['ha_port']
            else:
                the_port = router['gw_port']
            found = False
            for i in range(len(the_port['subnets'])):
                gw_ip = the_port['subnets'][i]['gateway_ip']
                if next_hop.lower() == gw_ip.lower():
                    found = True
                    break
            if found is False:
                LOG.info("route has incorrect next-hop, deleting")
                delete_route_list.append(route.text)
                continue

            self.existing_cfg_dict['routes'][router_id] = route

        if not self.test_mode:
            for route_cfg in delete_route_list:
                del_cmd = XML_CMD_TAG % ("no %s" % (route_cfg))
                confstr = XML_FREEFORM_SNIPPET % (del_cmd)
                LOG.info("Delete default route: %(del_cmd)s" %
                         {'del_cmd': del_cmd})
                conn.edit_config(target='running', config=confstr)

        LOG.debug("delete_route_list = %s " % (pp.pformat(delete_route_list)))
        return delete_route_list

    def clean_snat(self, conn, router_id_dict,
                   intf_segment_dict, segment_nat_dict, parsed_cfg):
        delete_fip_list = []
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        if (is_multi_region_enabled):
            snat_regex_old = SNAT_MULTI_REGION_REGEX_OLD
            snat_regex = SNAT_MULTI_REGION_REGEX
        else:
            snat_regex_old = SNAT_REGEX_OLD
            snat_regex = SNAT_REGEX

        # Delete any entries with old style 'hsrp-grp-x-y' grp name
        floating_ip_old_rules = parsed_cfg.find_objects(snat_regex_old)
        for snat_rule in floating_ip_old_rules:
            LOG.info("\n Rule is old format, deleting: %(snat_rule)s" %
                     {'snat_rule': snat_rule.text})

            delete_fip_list.append(snat_rule.text)

        floating_ip_nats = parsed_cfg.find_objects(snat_regex)
        for snat_rule in floating_ip_nats:
            LOG.info("\nstatic nat rule: %(snat_rule)s" %
                     {'snat_rule': snat_rule})
            match_obj = re.match(snat_regex, snat_rule.text)
            if is_multi_region_enabled:
                (inner_ip, outer_ip, router_id, region_id,
                    hsrp_num, segment_id) = match_obj.group(1, 2, 3, 4, 5, 6)
                my_region_id = cfg.CONF.multi_region.region_id
                other_region_ids = cfg.CONF.multi_region.other_region_ids
                if region_id != my_region_id:
                    if region_id not in other_region_ids:
                        delete_fip_list.append(snat_rule.text)
                    else:
                        # skip because some other deployment owns
                        # this configuration
                        continue
            else:
                inner_ip, outer_ip, router_id, hsrp_num, segment_id = (
                    match_obj.group(1, 2, 3, 4, 5))

            segment_id = int(segment_id)
            hsrp_num = int(hsrp_num)

            LOG.info("   in_ip: %(inner_ip)s, out_ip: %(outer_ip)s, "
                     "router_id: %(router_id)s, hsrp_num: %(hsrp_num)s,"
                     " segment_id: %(segment_id)s" %
                     {'inner_ip': inner_ip,
                      'outer_ip': outer_ip,
                      'router_id': router_id,
                      'hsrp_num': hsrp_num,
                      'segment_id': segment_id})

            # Check that VRF exists in openstack DB info
            if router_id not in router_id_dict:
                LOG.info("router not found for rule, deleting")
                delete_fip_list.append(snat_rule.text)
                continue

            # Check that router has external network and segment_id matches
            router = router_id_dict[router_id]
            if "gw_port" not in router:
                LOG.info("router has no gw_port,"
                         " snat is invalid, deleting")
                delete_fip_list.append(snat_rule.text)
                continue

            # Check that hsrp group name is correct
            gw_port = router['gw_port']
            gw_hsrp_num = int(gw_port[ha.HA_INFO]['group'])
            gw_segment_id = int(gw_port['hosting_info']['segmentation_id'])
            if segment_id != gw_segment_id:
                LOG.info("snat segment_id does not match router's"
                         " gw segment_id, deleting")
                delete_fip_list.append(snat_rule.text)
                continue

            if hsrp_num != gw_hsrp_num:
                LOG.info("snat hsrp group does not match router gateway's"
                         " hsrp group, deleting")
                delete_fip_list.append(snat_rule.text)
                continue

            # Check that in,out ip pair matches a floating_ip defined on router
            if '_floatingips' not in router:
                LOG.info("Router has no floating IPs defined,"
                         " snat rule is invalid, deleting")
                delete_fip_list.append(snat_rule.text)
                continue

            fip_match_found = False
            for floating_ip in router['_floatingips']:
                if inner_ip == floating_ip['fixed_ip_address'] and \
                   outer_ip == floating_ip['floating_ip_address']:
                    fip_match_found = True
                    break
            if fip_match_found is False:
                LOG.info("snat rule does not match defined floating IPs,"
                         " deleting")
                delete_fip_list.append(snat_rule.text)
                continue

            self.existing_cfg_dict['static_nat'][outer_ip] = snat_rule

        if not self.test_mode:
            for fip_cfg in delete_fip_list:
                del_cmd = XML_CMD_TAG % ("no %s" % (fip_cfg))
                confstr = XML_FREEFORM_SNIPPET % (del_cmd)
                LOG.info("Delete SNAT: %(del_cmd)s" %
                         {'del_cmd': del_cmd})
                conn.edit_config(target='running', config=confstr)

        LOG.debug("delete_fip_list = %s " % (pp.pformat(delete_fip_list)))
        return delete_fip_list

    def clean_nat_pool_overload(self,
                                conn,
                                router_id_dict,
                                intf_segment_dict,
                                segment_nat_dict,
                                parsed_cfg):
        delete_nat_list = []

        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        if (is_multi_region_enabled):
            nat_pool_overload_regex = NAT_POOL_OVERLOAD_MULTI_REGION_REGEX
        else:
            nat_pool_overload_regex = NAT_POOL_OVERLOAD_REGEX

        nat_overloads = parsed_cfg.find_objects(nat_pool_overload_regex)
        for nat_rule in nat_overloads:
            LOG.info("\nnat overload rule: %(nat_rule)s" %
                     {'nat_rule': nat_rule})
            match_obj = re.match(nat_pool_overload_regex, nat_rule.text)
            if is_multi_region_enabled:
                (region_id, segment_id, port_id,
                 pool_router_id, pool_region_id, router_id) = (
                    match_obj.group(1, 2, 3, 4, 5, 6))
                if (region_id != pool_region_id):
                    LOG.info("region id mismatch, deleting")
                    delete_nat_list.append(nat_rule.text)
                    continue

                my_region_id = cfg.CONF.multi_region.region_id
                other_region_ids = cfg.CONF.multi_region.other_region_ids
                if (region_id != my_region_id):
                    if region_id not in other_region_ids:
                        delete_nat_list.append(nat_rule.text)
                    else:
                        # skip because some other deployment owns
                        # this configuration
                        continue
            else:
                segment_id, pool_router_id, router_id = (
                    match_obj.group(1, 2, 3))
                segment_id, port_id, pool_router_id, router_id = (
                    match_obj.group(1, 2, 3, 4))

            segment_id = int(segment_id)

            # Check that VRF exists in openstack DB info
            if router_id not in router_id_dict:
                LOG.info("router not found for rule, deleting")
                delete_nat_list.append(nat_rule.text)
                continue

            # Check that correct pool is specified
            if pool_router_id != router_id:
                LOG.info("Pool and VRF name mismatch, deleting")
                delete_nat_list.append(nat_rule.text)
                continue

            # Check that router has external network
            router = router_id_dict[router_id]
            if "gw_port" not in router:
                LOG.info("router has no gw_port,"
                         " nat overload is invalid, deleting")
                delete_nat_list.append(nat_rule.text)
                continue

            # Check that router has internal network interface on segment_id
            intf_match_found = False
            if '_interfaces' in router:
                for intf in router['_interfaces']:
                    if intf['device_owner'] == \
                            bc.constants.DEVICE_OWNER_ROUTER_INTF:
                        intf_segment_id = \
                            int(intf['hosting_info']['segmentation_id'])
                        if intf_segment_id == segment_id:
                            intf_match_found = True
                            break

            if intf_match_found is False:
                LOG.info("router does not have this internal network"
                         " assigned, deleting rule")
                delete_nat_list.append(nat_rule.text)
                continue

            self.existing_cfg_dict['dyn_nat'][segment_id] = nat_rule

        if not self.test_mode:
            for nat_cfg in delete_nat_list:
                del_cmd = XML_CMD_TAG % ("no %s" % (nat_cfg))
                confstr = XML_FREEFORM_SNIPPET % (del_cmd)
                LOG.info("Delete NAT overload: %(del_cmd)s" %
                         {'del_cmd': del_cmd})
                conn.edit_config(target='running', config=confstr)

        LOG.debug("delete_nat_list = %s " % (pp.pformat(delete_nat_list)))
        return delete_nat_list

    def check_acl_permit_rules_valid(self, segment_id, acl, intf_segment_dict):
        permit_rules = acl.re_search_children(ACL_CHILD_REGEX)
        for permit_rule in permit_rules:
            LOG.info("   permit rule: %(permit_rule)s" %
                     {'permit_rule': permit_rule})
            match_obj = re.match(ACL_CHILD_REGEX, permit_rule.text)
            net_ip, hostmask = match_obj.group(1, 2)

            cfg_subnet = netaddr.IPNetwork("%s/%s" % (net_ip, hostmask))

            db_subnet = netaddr.IPNetwork("255.255.255.255/32")  # dummy value
            try:
                intf_list = intf_segment_dict[segment_id]
                for intf in intf_list:
                    if intf['device_owner'] == \
                            bc.constants.DEVICE_OWNER_ROUTER_INTF:
                        subnet_cidr = intf['subnets'][0]['cidr']
                        db_subnet = netaddr.IPNetwork(subnet_cidr)
                        break
            except KeyError:
                LOG.info("KeyError when attemping to validate segment_id")
                return False

            LOG.info("   cfg_subnet: %(cfg_net)s/%(cfg_prefix_len)s,"
                     " db_subnet: %(db_net)s/%(db_prefix_len)s" %
                     {'cfg_net': cfg_subnet.network,
                      'cfg_prefix_len': cfg_subnet.prefixlen,
                      'db_net': db_subnet.network,
                      'db_prefix_len': db_subnet.prefixlen})

            if cfg_subnet.network != db_subnet.network or \
               cfg_subnet.prefixlen != db_subnet.prefixlen:
                LOG.info("ACL subnet does not match subnet info"
                         " in openstack DB, deleting ACL")
                return False

        return True

    def clean_acls(self, conn, intf_segment_dict,
                   segment_nat_dict, parsed_cfg):

        delete_acl_list = []
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        if (is_multi_region_enabled):
            acl_regex = ACL_MULTI_REGION_REGEX
        else:
            acl_regex = ACL_REGEX

        acls = parsed_cfg.find_objects(acl_regex)
        for acl in acls:
            LOG.info("\nacl: %(acl)s" % {'acl': acl})
            match_obj = re.match(acl_regex, acl.text)

            if (is_multi_region_enabled):
                region_id = match_obj.group(1)
                segment_id = int(match_obj.group(2))
                port_id = match_obj.group(3)
                if region_id != cfg.CONF.multi_region.region_id:
                    if region_id not in cfg.CONF.multi_region.other_region_ids:
                        delete_acl_list.append(acl.text)
                    else:
                        # skip because some other deployment owns
                        # this configuration
                        continue
            else:
                segment_id = int(match_obj.group(1))
                port_id = match_obj.group(2)

            LOG.info("   segment_id: %(seg_id)s, port_id: %(port_id)s" %
                     {'seg_id': segment_id, 'port_id': port_id})

            # Check that segment_id exists in openstack DB info
            if segment_id not in intf_segment_dict:
                LOG.info("Segment ID not found, deleting acl")
                delete_acl_list.append(acl.text)
                continue

            # Check that permit rules match subnets defined on openstack intfs
            if self.check_acl_permit_rules_valid(segment_id,
                                                 acl,
                                                 intf_segment_dict) is False:
                delete_acl_list.append(acl.text)
                continue

            self.existing_cfg_dict['acls'][segment_id] = acl

        if not self.test_mode:
            for acl_cfg in delete_acl_list:
                del_cmd = XML_CMD_TAG % ("no %s" % (acl_cfg))
                confstr = XML_FREEFORM_SNIPPET % (del_cmd)
                LOG.info("Delete ACL: %(del_cmd)s" % {'del_cmd': del_cmd})
                conn.edit_config(target='running', config=confstr)

        LOG.debug("delete_acl_list = %s" % (pp.pformat(delete_acl_list)))
        return delete_acl_list

    def subintf_real_ip_check_gw_port(self, gw_port, ip_addr, netmask):
        """
        checks running-cfg derived ip_addr and netmask against neutron-db
        gw_port
        """
        if gw_port is not None:
            found = False
            for i in range(len(gw_port['fixed_ips'])):
                target_ip = gw_port['fixed_ips'][i]['ip_address']
                if ip_addr == target_ip:
                    found = True
                    break
            if found is False:
                LOG.info("Subintf real IP is incorrect, deleting")
                return False
            subnet_id = gw_port['fixed_ips'][i]['subnet_id']
            subnet = next(
                sn for sn in gw_port['subnets'] if sn['id'] == subnet_id)
            target_net = netaddr.IPNetwork(subnet['cidr'])
            if netmask != str(target_net.netmask):
                LOG.info("Subintf has incorrect netmask, deleting")
                return False

            return True

        return False

    def subintf_real_ip_check(self, intf_list, ip_addr, netmask):

        for target_intf in intf_list:
            target_ip = target_intf['fixed_ips'][0]['ip_address']
            target_net = netaddr.IPNetwork(target_intf['subnets'][0]['cidr'])
            LOG.info("target ip,net: %(target_ip)s,%(target_net)s,"
                     " actual ip,net %(ip_addr)s,%(netmask)s" %
                     {'target_ip': target_ip,
                      'target_net': target_net,
                      'ip_addr': ip_addr,
                      'netmask': netmask})

            if ip_addr != target_ip:
                LOG.info("Subintf real IP is incorrect, deleting")
                return False
            if netmask != str(target_net.netmask):
                LOG.info("Subintf has incorrect netmask, deleting")
                return False

            return True

        return False

    def subintf_real_ipv6_check(self,
                                intf_list,
                                is_external,
                                ipv6_addr,
                                prefixlen):

        if is_external:
            target_type = bc.constants.DEVICE_OWNER_ROUTER_GW
        else:
            target_type = bc.constants.DEVICE_OWNER_ROUTER_INTF

        for target_intf in intf_list:
            if target_intf['device_owner'] == target_type:

                target_ip = \
                    netaddr.IPAddress(
                        target_intf['fixed_ips'][0]['ip_address'])

                target_prefixlen = \
                    netaddr.IPNetwork(
                        target_intf['subnets'][0]['cidr']).prefixlen

                LOG.info("target ip,prefixlen: %(target_ip)s,"
                         "%(target_prefixlen)s, actual ip,"
                         "prefixlen %(ipv6_addr)s,%(prefixlen)s" %
                         {'target_ip': target_ip,
                          'target_prefixlen': target_prefixlen,
                          'ipv6_addr': ipv6_addr,
                          'prefixlen': prefixlen})

                if target_ip != netaddr.IPAddress(ipv6_addr):
                    LOG.info("Subintf IPv6 addr is incorrect, deleting")
                    return False

                if target_prefixlen != int(prefixlen):
                    LOG.info("Subintf IPv6 prefix length"
                             " is incorrect, deleting")
                    return False

                return True

        return False

    def gw_port_hsrp_ip_check(self, gw_port, ip_addr):

        if gw_port is not None:
            ha_port = gw_port[ha.HA_INFO]['ha_port']
            for fixed_ip in ha_port['fixed_ips']:
                target_ip = fixed_ip['ip_address']
                LOG.info("target_ip: %(target_ip)s, "
                         "actual_ip: %(ip_addr)s" %
                         {'target_ip': target_ip, 'ip_addr': ip_addr})
                if ip_addr == target_ip:
                    return True
            LOG.info("HSRP VIP mismatch on gw_port, deleting")
        return False

    def subintf_hsrp_ip_check(self, intf_list, is_external, ip_addr):
        for target_intf in intf_list:
            ha_intf = target_intf[ha.HA_INFO]['ha_port']
            target_ip = ha_intf['fixed_ips'][0]['ip_address']
            LOG.info("target_ip: %(target_ip)s, actual_ip: %(ip_addr)s" %
                     {'target_ip': target_ip,
                      'ip_addr': ip_addr})
            if ip_addr != target_ip:
                LOG.info("HSRP VIP mismatch, deleting")
                return False

            return True

        return False

    # Returns True if interface has correct NAT config
    def clean_interfaces_nat_check(self, intf, segment_nat_dict):
        intf_nat_type = intf.re_search_children(INTF_NAT_REGEX)
        intf_nat_type = self.get_single_cfg(intf_nat_type)

        if intf_nat_type is not None:
            intf_nat_type = intf_nat_type.re_match(INTF_NAT_REGEX, group=1)

        LOG.info("NAT Type: %s", intf_nat_type)

        intf.nat_type = intf_nat_type

        if (intf.segment_id in segment_nat_dict and
            segment_nat_dict[intf.segment_id] is True):
            if intf.is_external:
                if intf_nat_type != "outside":
                    nat_cmd = XML_CMD_TAG % (intf.text)
                    nat_cmd += XML_CMD_TAG % ("ip nat outside")
                    confstr = XML_FREEFORM_SNIPPET % (nat_cmd)
                    LOG.info("NAT type mismatch, should be outside")
                    return False
            else:
                if intf_nat_type != "inside":
                    nat_cmd = XML_CMD_TAG % (intf.text)
                    nat_cmd += XML_CMD_TAG % ("ip nat inside")
                    confstr = XML_FREEFORM_SNIPPET % (nat_cmd)
                    LOG.info("NAT type mismatch, should be inside")
                    return False
        else:
            if intf_nat_type is not None:
                nat_cmd = XML_CMD_TAG % (intf.text)
                nat_cmd += XML_CMD_TAG % ("no ip nat %s" % (intf_nat_type))
                confstr = XML_FREEFORM_SNIPPET % (nat_cmd)
                LOG.info("NAT type mismatch, should have no NAT, %s",
                         (confstr))
                return False

        return True

    def clean_interfaces_ipv4_hsrp_check(self, intf, intf_db_dict):
        # Check HSRP VIP
        hsrp_vip_cfg_list = intf.re_search_children(HSRP_V4_VIP_REGEX)
        num_vips = len(hsrp_vip_cfg_list)
        num_vips_db = len(intf_db_dict[intf.segment_id][0]['fixed_ips'])
        if num_vips < 1:
            LOG.info("Interface is missing HSRP VIP, deleting")
            return False
        elif num_vips != num_vips_db:
            LOG.info("Subintf has wrong number of HSRP VIP addresses ("
                     "should have %(n_v)d, but has %(n_v_d)d), deleting",
                     {'n_v': num_vips, 'n_v_d': num_vips_db})
            return False

        if intf.is_external:
            for hsrp_vip_cfg in hsrp_vip_cfg_list:
                match_obj = re.match(HSRP_V4_VIP_REGEX, hsrp_vip_cfg.text)
                hsrp_vip_grp_num, hsrp_vip = match_obj.group(1, 2)
                if self.gw_port_hsrp_ip_check(intf_db_dict[intf.segment_id][0],
                                              hsrp_vip) is False:
                    return False
            return True
        else:
            hsrp_vip_cfg = hsrp_vip_cfg_list[0]
            match_obj = re.match(HSRP_V4_VIP_REGEX, hsrp_vip_cfg.text)
            hsrp_vip_grp_num, hsrp_vip = match_obj.group(1, 2)
            return self.subintf_hsrp_ip_check(
                                            intf_db_dict[intf.segment_id],
                                            intf.is_external,
                                            hsrp_vip)

    def clean_interfaces_ipv4_check(self, intf, intf_db_dict):

        # Check that real IP address is correct
        ipv4_addr = intf.re_search_children(INTF_V4_ADDR_REGEX)
        num_addrs = len(ipv4_addr)
        num_addrs_db = len(intf_db_dict[intf.segment_id][0]['fixed_ips'])
        if num_addrs < 1:
            LOG.info("Subintf has no IP address, deleting")
            return False
        elif num_addrs != num_addrs_db:
            LOG.info("Subintf has wrong number of addresses (should have "
                     "%(n_a)d, but has %(n_a_d)d), deleting", num_addrs,
                     num_addrs_db)
            return False

        if intf.is_external:
            for ipv4_addr_cfg in ipv4_addr:
                match_obj = re.match(INTF_V4_ADDR_REGEX, ipv4_addr_cfg.text)
                ip_addr, netmask = match_obj.group(1, 2)
                if (self.subintf_real_ip_check_gw_port(
                        intf_db_dict[intf.segment_id][0], ip_addr, netmask) is
                        False):
                    return False
            return True
        else:
            ipv4_addr_cfg = ipv4_addr[0]
            match_obj = re.match(INTF_V4_ADDR_REGEX, ipv4_addr_cfg.text)
            ip_addr, netmask = match_obj.group(1, 2)
            return self.subintf_real_ip_check(
                intf_db_dict[intf.segment_id], ip_addr, netmask)

    def clean_interfaces_ipv6_check(self, intf, intf_segment_dict):
        # Check that real IP address is correct
        ipv6_addr = intf.re_search_children(INTF_V6_ADDR_REGEX)
        if len(ipv6_addr) < 1:
            LOG.info("Subintf has no IPv6 address, deleting")
            return False

        ipv6_addr_cfg = ipv6_addr[0]
        match_obj = re.match(INTF_V6_ADDR_REGEX, ipv6_addr_cfg.text)
        ipv6_addr, prefixlen = match_obj.group(1, 2)

        return self.subintf_real_ipv6_check(intf_segment_dict[intf.segment_id],
                                            intf.is_external,
                                            ipv6_addr, prefixlen)

    def clean_interfaces(self, conn, intf_segment_dict,
                         segment_nat_dict, parsed_cfg):
        is_multi_region_enabled = cfg.CONF.multi_region.enable_multi_region
        if (is_multi_region_enabled):
            intf_desc_regex = INTF_DESC_MULTI_REGION_REGEX
            vrf_intf_regex_new = VRF_INTF_MULTI_REGION_REGEX_NEW
        else:
            intf_desc_regex = INTF_DESC_REGEX
            vrf_intf_regex_new = VRF_INTF_REGEX_NEW
        runcfg_intfs = [obj for obj in parsed_cfg.find_objects("^interf")
                        if obj.re_search_children(intf_desc_regex)]

        pending_delete_list = []

        # TODO(nh0556) split this big function into smaller functions
        for intf in runcfg_intfs:
            LOG.info("ASR interface: %s", (intf))

            # check for region-id
            if is_multi_region_enabled:
                intf_desc = intf.re_search_children(intf_desc_regex)
                intf_desc = self.get_single_cfg(intf_desc)

                if not intf_desc:
                    LOG.info("Interface doesn't have "
                             "a interface description, ignoring %s" % (
                                 intf.text))
                    continue
                else:
                    intf_desc_region_id = (
                        intf_desc.re_match(INTF_DESC_MULTI_REGION_REGEX,
                                           group=1))
                    if (intf_desc_region_id is None):
                        # if no matching description field is found,
                        # assume that the interface is not owned by Openstack
                        LOG.info("Interface description doesn't have"
                                 "region-id, ignoring %s" % intf.text)
                        continue
                    else:
                        # since region-id was found, check if it's valid
                        my_region_id = cfg.CONF.multi_region.region_id
                        other_region_ids = (
                            cfg.CONF.multi_region.other_region_ids)
                        if intf_desc_region_id != my_region_id:
                            if (intf_desc_region_id not in other_region_ids):
                                pending_delete_list.append(intf)
                            else:
                                # skip because some other deployment owns
                                # this configuration
                                continue

            # obtain segment_id from interface definition
            intf.segment_id = int(intf.re_match(INTF_REGEX, group=1))
            LOG.info("  segment_id: %s", (intf.segment_id))

            # Delete any interfaces where config doesn't match DB
            # Correct config will be added after clearing invalid cfg

            # TODO(Check that interface name e.g. Port-channel10 matches)
            # TODO(that specified in .ini file)

            # Check that the interface segment_id exists in the current DB data
            if (intf.segment_id not in intf_segment_dict and
               intf.segment_id not in self.segment_gw_dict):
                LOG.info("Invalid segment ID, delete interface")
                pending_delete_list.append(intf)
                continue

            if (intf.segment_id in self.segment_gw_dict):
                intf.is_external = True
            else:
                intf.is_external = False

            # Check if dot1q config is correct
            dot1q_cfg = intf.re_search_children(DOT1Q_REGEX)
            dot1q_cfg = self.get_single_cfg(dot1q_cfg)

            if dot1q_cfg is None:
                LOG.info("Missing DOT1Q config, delete interface")
                pending_delete_list.append(intf)
                continue
            else:
                dot1q_num = int(dot1q_cfg.re_match(DOT1Q_REGEX, group=1))
                if dot1q_num != int(intf.segment_id):
                    LOG.info("DOT1Q mismatch, delete interface")
                    pending_delete_list.append(intf)
                    continue

            # Is this an "external network" segment_id?
            if intf.is_external:
                db_intf = self.segment_gw_dict[intf.segment_id][0]
            else:
                db_intf = intf_segment_dict[intf.segment_id][0]

            intf.has_ipv6 = is_port_v6(db_intf)

            # Check VRF config
            if intf.is_external:
                vrf_cfg = intf.re_search_children(VRF_EXT_INTF_REGEX_NEW)
                vrf_cfg = self.get_single_cfg(vrf_cfg)
                LOG.info("VRF: %s", (vrf_cfg))

                if vrf_cfg is not None:  # external network has no vrf
                    LOG.info("External network shouldn't"
                             " have VRF, deleting intf")
                    pending_delete_list.append(intf)
                    continue
            else:
                vrf_cfg = intf.re_search_children(vrf_intf_regex_new)
                vrf_cfg = self.get_single_cfg(vrf_cfg)
                LOG.info("VRF: %s", (vrf_cfg))
                if not vrf_cfg:
                    LOG.info("Internal network missing valid VRF,"
                             " deleting intf")
                    pending_delete_list.append(intf)
                    continue

                # check for VRF mismatch
                match_obj = re.match(vrf_intf_regex_new, vrf_cfg.text)
                router_id = match_obj.group(1)
                if (is_multi_region_enabled):
                    region_id = match_obj.group(2)
                    my_region_id = cfg.CONF.multi_region.region_id
                    other_region_ids = cfg.CONF.multi_region.other_region_ids
                    if region_id != my_region_id:
                        if (region_id not in other_region_ids):
                            pending_delete_list.append(intf)
                        else:
                            # skip because some other deployment owns
                            # this configuration
                            continue

                # router_id device_id/ha_port_device_id check
                if (ha.HA_INFO in db_intf):
                    ha_port_device_id = (db_intf[ha.HA_INFO]
                                         ['ha_port']['device_id'])
                else:
                    ha_port_device_id = None

                if (router_id != db_intf["device_id"][0:6] and
                    (ha_port_device_id is not None and
                     router_id != ha_port_device_id[0:6])):

                    LOG.info("Internal network VRF mismatch,"
                             " deleting intf,"
                             " router_id: %(router_id)s,"
                             " db_intf_dev_id: %(db_intf_dev_id)s" %
                             {'router_id': router_id,
                              'db_intf_dev_id': db_intf["device_id"]})
                    pending_delete_list.append(intf)
                    continue

            correct_grp_num = int(db_intf[ha.HA_INFO]['group'])

            if intf.is_external:
                intf_db = self.segment_gw_dict
            else:
                intf_db = intf_segment_dict

            if intf.has_ipv6 is False:
                if self.clean_interfaces_nat_check(intf,
                                                   segment_nat_dict) is False:
                    pending_delete_list.append(intf)
                    continue
                if self.clean_interfaces_ipv4_check(intf, intf_db) is False:
                    pending_delete_list.append(intf)
                    continue
                if self.clean_interfaces_ipv4_hsrp_check(intf,
                                                         intf_db) is False:
                    pending_delete_list.append(intf)
                    continue
            else:
                if self.clean_interfaces_ipv6_check(intf, intf_db) is False:
                    pending_delete_list.append(intf)
                    continue

            # Delete if there's any hsrp config with wrong group number
            hsrp_cfg_list = intf.re_search_children(HSRP_REGEX)
            needs_hsrp_delete = False
            for hsrp_cfg in hsrp_cfg_list:
                hsrp_num = int(hsrp_cfg.re_match(HSRP_REGEX, group=1))
                if hsrp_num != correct_grp_num:
                    needs_hsrp_delete = True

            if needs_hsrp_delete:
                LOG.info("Bad HSRP config for interface, deleting")
                pending_delete_list.append(intf)
                continue

            self.existing_cfg_dict['interfaces'][intf.segment_id] = intf.text

        if not self.test_mode:
            for intf in pending_delete_list:
                del_cmd = XML_CMD_TAG % ("no %s" % (intf.text))
                confstr = XML_FREEFORM_SNIPPET % (del_cmd)
                LOG.info("Deleting %s", (intf.text))
                conn.edit_config(target='running', config=confstr)

        LOG.debug("pending_delete_list (interfaces) = %s" %
                  pp.pformat(pending_delete_list))
        return pending_delete_list



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import mock

from neutron.tests import base

from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.server import cisco_dfa_rest as dc

"""This file includes test cases for cisco_dfa_rest.py."""

FAKE_DCNM_IP = '1.1.1.1'
FAKE_DCNM_USERNAME = 'dcnmuser'
FAKE_DCNM_PASSWD = 'dcnmpass'


class TestNetwork(object):
    segmentation_id = 123456
    name = 'cisco_test_network'
    config_profile = 'defaultL2ConfigProfile'


class TestCiscoDFAClient(base.BaseTestCase):
    """Test cases for DFARESTClient."""

    def setUp(self):
        # Declare the test resource.
        super(TestCiscoDFAClient, self).setUp()

        # Setting DCNM parameters.
        config.default_dcnm_opts['dcnm']['dcnm_ip'] = FAKE_DCNM_IP
        config.default_dcnm_opts['dcnm']['dcnm_user'] = FAKE_DCNM_USERNAME
        config.default_dcnm_opts['dcnm']['dcnm_password'] = FAKE_DCNM_PASSWD
        config.default_dcnm_opts['dcnm']['timeout_resp'] = 0.01
        self.cfg = config.CiscoDFAConfig().cfg

        self.send_req = mock.patch.object(dc.DFARESTClient,
                                          '_send_request').start()
        self.get_vers = mock.patch.object(dc.DFARESTClient,
                                          'get_version').start()
        self.login = mock.patch.object(dc.DFARESTClient,
                                       '_login_request').start()
        self.logout = mock.patch.object(dc.DFARESTClient,
                                        '_logout_request').start()
        self.send_req.return_value = mock.Mock()
        self.send_req.return_value.status_code = 200
        self.send_req.return_value.json.return_value = {}
        self.get_vers.return_value = '7.0(1)'

        self.dcnm_client = dc.DFARESTClient(self.cfg)
        self.testnetwork = TestNetwork()

    def test_create_project(self):
        """Test create project."""
        self.send_req.reset_mock()

        org_name = 'Cisco'
        part_name = self.dcnm_client._part_name
        dci = 100
        orch_id = 'OpenStack Controller'
        self.dcnm_client.create_project(orch_id, org_name, part_name, dci)
        call_cnt = self.dcnm_client._send_request.call_count
        self.assertEqual(2, call_cnt)

        # Check call to partition and organization happens.
        org_pyld = {
            'organizationName': org_name,
            'description': org_name,
            'orchestrationSource': "OpenStack Controller"}
        part_pyld = {
            'partitionName': part_name,
            'organizationName': org_name,
            'description': org_name,
            'serviceNodeIpAddress': None}
        org_url = self.dcnm_client._org_url
        part_url = self.dcnm_client._create_part_url % org_name
        expected_calls = [mock.call('POST', org_url, org_pyld, 'organization'),
                          mock.call('POST', part_url, part_pyld, 'partition')]
        self.assertEqual(expected_calls,
                         self.dcnm_client._send_request.call_args_list)

    def test_create_network(self):
        """Test create network."""

        network_info = {}
        cfg_args = []
        seg_id = str(self.testnetwork.segmentation_id)
        config_profile = self.testnetwork.config_profile
        network_name = self.testnetwork.name
        tenant_name = 'Cisco'
        part_name = self.dcnm_client._part_name
        url = self.dcnm_client._create_network_url % (tenant_name, part_name)

        cfg_args.append("$segmentId=" + seg_id)
        cfg_args.append("$netMaskLength=16")
        cfg_args.append("$gatewayIpAddress=30.31.32.1")
        cfg_args.append("$networkName=" + network_name)
        cfg_args.append("$vlanId=0")
        cfg_args.append("$vrfName=%s:%s" % (tenant_name, part_name))
        cfg_args = ';'.join(cfg_args)

        dhcp_scopes = {'ipRange': '10.11.12.14-10.11.12.254',
                       'subnet': '10.11.12.13',
                       'gateway': '10.11.12.1'}

        network_info = {"segmentId": seg_id,
                        "vlanId": "0",
                        "mobilityDomainId": "None",
                        "profileName": config_profile,
                        "networkName": network_name,
                        "configArg": cfg_args,
                        "organizationName": tenant_name,
                        "partitionName": part_name,
                        "description": network_name,
                        "dhcpScope": dhcp_scopes}

        self.dcnm_client._create_network(network_info)
        self.dcnm_client._send_request.assert_called_with('POST', url,
                                                          network_info,
                                                          'network')

    def test_delete_network(self):
        """Test delete network."""

        seg_id = self.testnetwork.segmentation_id
        tenant_name = 'cisco'
        part_name = self.dcnm_client._part_name
        url = self.dcnm_client._network_url % (tenant_name, part_name, seg_id)
        self.dcnm_client.delete_network(tenant_name, self.testnetwork)
        self.dcnm_client._send_request.assert_called_with('DELETE', url,
                                                          '', 'network')

    def test_delete_project(self):
        """Test delete tenant."""
        self.send_req.reset_mock()

        tenant_name = 'cisco'
        part_name = self.dcnm_client._part_name
        self.dcnm_client.delete_project(tenant_name, part_name)
        call_cnt = self.dcnm_client._send_request.call_count
        self.assertEqual(2, call_cnt)

        # Check the calls to delete partition and org happens.
        del_org_url = self.dcnm_client._del_org_url % tenant_name
        del_part_url = self.dcnm_client._del_part % (tenant_name, part_name)
        expected_calls = [mock.call('DELETE', del_part_url, '', 'partition'),
                          mock.call('DELETE', del_org_url, '', 'organization')]
        self.assertEqual(expected_calls,
                         self.dcnm_client._send_request.call_args_list)

    def test_http_verify_protocol(self):
        """Test login test using http. """

        self.login.reset_mock()
        self.logout.reset_mock()
        self.dcnm_client._verify_protocol('http')
        test_login_url = 'http://' + FAKE_DCNM_IP + '/rest/logon'
        test_logout_url = 'http://' + FAKE_DCNM_IP + '/rest/logout'
        self.login.assert_called_with(test_login_url)
        self.logout.assert_called_with(test_logout_url)

    def test_https_verify_protocol(self):
        """Test login test using https. """

        self.login.reset_mock()
        self.logout.reset_mock()
        self.dcnm_client._verify_protocol('https')
        test_login_url = 'https://' + FAKE_DCNM_IP + '/rest/logon'
        test_logout_url = 'https://' + FAKE_DCNM_IP + '/rest/logout'
        self.login.assert_called_with(test_login_url)
        self.logout.assert_called_with(test_logout_url)

    def test_get_segmentid_range(self):
        """Test get segment ID range."""

        self.send_req.reset_mock()
        orch_id = 'OpenStack_Controller'
        self.dcnm_client.get_segmentid_range(orch_id)

        segment_range_url = self.dcnm_client._segmentid_ranges_url + '/' + (
            orch_id)
        expected_calls = [mock.call('GET', segment_range_url, None,
                                    'segment-id range')]
        self.assertEqual(expected_calls,
                         self.dcnm_client._send_request.call_args_list)

    def test_set_segmentid_range(self):
        """Test set segment ID range."""

        self.send_req.reset_mock()
        orch_id = 'OpenStack_Controller'
        segid_min = 10000
        segid_max = 12000
        self.dcnm_client.set_segmentid_range(orch_id, segid_min, segid_max)

        segment_range_url = self.dcnm_client._segmentid_ranges_url
        payload = {'orchestratorId': orch_id,
                   'segmentIdRanges': "%s-%s" % (segid_min, segid_max)}

        expected_calls = [mock.call('POST', segment_range_url, payload,
                                    'segment-id range')]
        self.assertEqual(expected_calls,
                         self.dcnm_client._send_request.call_args_list)

    def test_update_segmentid_range(self):
        """Test set segment ID range."""

        self.send_req.reset_mock()
        orch_id = 'OpenStack_Controller'
        segid_min = 10000
        segid_max = 12000
        self.dcnm_client.update_segmentid_range(orch_id, segid_min, segid_max)

        segment_range_url = self.dcnm_client._segmentid_ranges_url + '/' + (
            orch_id)
        payload = {'orchestratorId': orch_id,
                   'segmentIdRanges': "%s-%s" % (segid_min, segid_max)}

        expected_calls = [mock.call('PUT', segment_range_url, payload,
                                    'segment-id range')]
        self.assertEqual(expected_calls,
                         self.dcnm_client._send_request.call_args_list)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from sqlalchemy import sql
from sqlalchemy.sql import expression as expr

from networking_cisco._i18n import _

from neutron.common import topics
from neutron.db import agents_db
from neutron.db import l3_agentschedulers_db
from neutron.db import models_v2
from neutron.db import portbindings_db as p_binding

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import l3_const
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.db.device_manager import hd_models
from networking_cisco.plugins.cisco.db.l3 import l3_models
from networking_cisco.plugins.cisco.extensions import routertypeawarescheduler

LOG = logging.getLogger(__name__)

AGENT_TYPE_CFG = cisco_constants.AGENT_TYPE_CFG
AGENT_TYPE_L3_CFG = cisco_constants.AGENT_TYPE_L3_CFG


ROUTER_TYPE_AWARE_SCHEDULER_OPTS = [
    cfg.StrOpt('router_type_aware_scheduler_driver',
               default='networking_cisco.plugins.cisco.l3.schedulers.'
                       'l3_routertype_aware_agent_scheduler.'
                       'L3RouterTypeAwareScheduler',
               help=_('Driver to use for router type-aware scheduling of '
                      'router to a default L3 agent')),
    cfg.BoolOpt('auto_schedule',
                default=True,
                help=_('Set to True if routers are to be automatically '
                       'scheduled by default')),
    cfg.BoolOpt('share_hosting_device',
                default=True,
                help=_('Set to True if routers can share hosts with routers '
                       'owned by other tenants by default')),

]

cfg.CONF.register_opts(ROUTER_TYPE_AWARE_SCHEDULER_OPTS, "routing")


class L3RouterTypeAwareSchedulerDbMixin(
        l3_agentschedulers_db.L3AgentSchedulerDbMixin):
    """Mixin class to add L3 router type-aware scheduler capability.

    This class can schedule Neutron routers to hosting devices
    and to L3 agents on network nodes.
    """
    def validate_hosting_device_router_combination(self, context, binding_info,
                                                   hosting_device_id):
        #TODO(bobmel): Perform proper hosting device validation
        target_hd_db = self._dev_mgr._get_hosting_device(context,
                                                         hosting_device_id)
        if target_hd_db.admin_state_up is False:
            raise routertypeawarescheduler.InvalidHostingDevice(
                hosting_device_id=hosting_device_id)
        rt_info = self.get_routertypes(
            context, fields=['id', 'slot_need'],
            filters={'template_id': [target_hd_db.template_id]})
        if not rt_info:
            raise routertypeawarescheduler.RouterHostingDeviceMismatch(
                router_type=binding_info.routertype_id,
                router_id=binding_info.router_id,
                hosting_device_id=hosting_device_id)
        return rt_info[0]

    def add_router_to_hosting_device(self, context, hosting_device_id,
                                     router_id):
        """Add a (non-hosted) router to a hosting device."""
        e_context = context.elevated()
        r_hd_binding_db = self._get_router_binding_info(e_context, router_id)
        if r_hd_binding_db.hosting_device_id:
            if r_hd_binding_db.hosting_device_id == hosting_device_id:
                return
            raise routertypeawarescheduler.RouterHostedByHostingDevice(
                router_id=router_id, hosting_device_id=hosting_device_id)
        rt_info = self.validate_hosting_device_router_combination(
            context, r_hd_binding_db, hosting_device_id)
        result = self.schedule_router_on_hosting_device(
            e_context, r_hd_binding_db, hosting_device_id,
            rt_info['slot_need'])
        if result:
            # refresh so that we get latest contents from DB
            e_context.session.expire(r_hd_binding_db)
            router = self.get_router(e_context, router_id)
            self.add_type_and_hosting_device_info(
                e_context, router, r_hd_binding_db, schedule=False)
            l3_cfg_notifier = self.agent_notifiers.get(AGENT_TYPE_L3_CFG)
            if l3_cfg_notifier:
                l3_cfg_notifier.router_added_to_hosting_device(context, router)
        else:
            raise routertypeawarescheduler.RouterSchedulingFailed(
                router_id=router_id, hosting_device_id=hosting_device_id)

    def remove_router_from_hosting_device(self, context, hosting_device_id,
                                          router_id):
        """Remove the router from hosting device.

        After removal, the router will be non-hosted until there is update
        which leads to re-schedule or be added to another hosting device
        manually.
        """
        e_context = context.elevated()
        r_hd_binding_db = self._get_router_binding_info(e_context, router_id)
        if r_hd_binding_db.hosting_device_id != hosting_device_id:
            raise routertypeawarescheduler.RouterNotHostedByHostingDevice(
                router_id=router_id, hosting_device_id=hosting_device_id)
        router = self.get_router(context, router_id)
        self.add_type_and_hosting_device_info(
            e_context, router, r_hd_binding_db, schedule=False)
        # conditionally remove router from backlog ensure it does not get
        # scheduled automatically
        self.remove_router_from_backlog(id)
        l3_cfg_notifier = self.agent_notifiers.get(AGENT_TYPE_L3_CFG)
        if l3_cfg_notifier:
            l3_cfg_notifier.router_removed_from_hosting_device(context, router)
        LOG.debug("Unscheduling router %s", r_hd_binding_db.router_id)
        self.unschedule_router_from_hosting_device(context, r_hd_binding_db)
        # now unbind the router from the hosting device
        with e_context.session.begin(subtransactions=True):
            r_hd_binding_db.hosting_device_id = None
            e_context.session.add(r_hd_binding_db)

    def list_routers_on_hosting_device(self, context, hosting_device_id):
        query = context.session.query(
            l3_models.RouterHostingDeviceBinding.router_id)
        query = query.filter(
            l3_models.RouterHostingDeviceBinding.hosting_device_id ==
            hosting_device_id)
        router_ids = [item[0] for item in query]
        if router_ids:
            return {'routers': self.get_sync_data_ext(context,
                                                      router_ids=router_ids)}
        else:
            return {'routers': []}

    def list_hosting_devices_hosting_router(self, context, router_id):
        query = context.session.query(
            l3_models.RouterHostingDeviceBinding.hosting_device_id)
        query = query.filter(l3_models.RouterHostingDeviceBinding.router_id ==
                             router_id)
        hd_ids = [item[0] for item in query]
        if hd_ids:
            return {'hosting_devices':
                    self._dev_mgr.get_hosting_devices(context,
                                                      filters={'id': hd_ids})}
        else:
            return {'hosting_devices': []}

    def cfg_list_router_ids_on_host(self, context, host, router_ids=None,
                                hosting_device_ids=None):
        agent = self._get_agent_by_type_and_host(context, AGENT_TYPE_CFG, host)
        if not agent.admin_state_up:
            return []
        query = context.session.query(
            l3_models.RouterHostingDeviceBinding.router_id)
        query = query.join(hd_models.HostingDevice)
        query = query.filter(hd_models.HostingDevice.cfg_agent_id == agent.id)
        if router_ids:
            query = query.filter(
                l3_models.RouterHostingDeviceBinding.router_id.in_(router_ids))
        if hosting_device_ids:
            query = query.filter(
                l3_models.RouterHostingDeviceBinding.hosting_device_id.in_(
                    hosting_device_ids))
        return [item[0] for item in query]

    def list_active_sync_routers_on_hosting_devices(self, context, host,
                                                    router_ids=None,
                                                    hosting_device_ids=None):
        router_ids = self.cfg_list_router_ids_on_host(context, host,
                                                      router_ids,
                                                      hosting_device_ids)
        if router_ids:
            return self.get_sync_data_ext(context, router_ids=router_ids)
        else:
            return []

    def _ensure_router_scheduling_compliant(self, router):
        auto_schedule = router.pop(routertypeawarescheduler.AUTO_SCHEDULE_ATTR,
                                   bc.constants.ATTR_NOT_SPECIFIED)
        if auto_schedule is bc.constants.ATTR_NOT_SPECIFIED:
            auto_schedule = cfg.CONF.routing.auto_schedule
        share_host = router.pop(routertypeawarescheduler.SHARE_HOST_ATTR,
                                bc.constants.ATTR_NOT_SPECIFIED)
        if share_host is bc.constants.ATTR_NOT_SPECIFIED:
            share_host = cfg.CONF.routing.share_hosting_device
        return auto_schedule, share_host

    @staticmethod
    @bc.extends([l3_const.ROUTERS])
    def _extend_router_dict_scheduling_info(router_res, router_db):
        router_res[routertypeawarescheduler.AUTO_SCHEDULE_ATTR] = (
            (router_db.hosting_info or {}).get('auto_schedule'))
        router_res[routertypeawarescheduler.SHARE_HOST_ATTR] = (
            (router_db.hosting_info or {}).get('share_hosting_device'))

    def get_active_routers_for_host(self, context, host):
        query = context.session.query(
            l3_models.RouterHostingDeviceBinding.router_id)
        query = query.join(
            models_v2.Port,
            l3_models.RouterHostingDeviceBinding.hosting_device_id ==
            models_v2.Port.device_id)
        query = query.join(p_binding.PortBindingPort)
        query = query.filter(p_binding.PortBindingPort.host == host)
        query = query.filter(models_v2.Port.name == 'mgmt')
        router_ids = [item[0] for item in query]
        return self.get_sync_data_ext(context, router_ids=router_ids,
                                      active=True)

    def _agent_state_filter(self, check_active, last_heartbeat):
        """Filters only active agents, if requested."""
        if not check_active:
            return True
        return not bc.is_agent_down(last_heartbeat)

    def get_hosts_for_routers(self, context, routers, admin_state_up=None,
                              check_active=False):
        query = context.session.query(p_binding.PortBindingPort.host,
                                      agents_db.Agent)
        query = query.join(
            models_v2.Port,
            models_v2.Port.id == p_binding.PortBindingPort.port_id)
        query = query.join(
            l3_models.RouterHostingDeviceBinding,
            l3_models.RouterHostingDeviceBinding.hosting_device_id ==
            models_v2.Port.device_id)
        query = query.join(
            agents_db.Agent,
            agents_db.Agent.host == p_binding.PortBindingPort.host)
        query = query.filter(sql.and_(
            agents_db.Agent.topic == topics.L3_AGENT,
            l3_models.RouterHostingDeviceBinding.router_id.in_(routers)))
        if admin_state_up is not None:
            query = query.filter(
                agents_db.Agent.admin_state_up == admin_state_up)
        hosts = [row.host for row in query if
                 self._agent_state_filter(check_active,
                                          row.heartbeat_timestamp)]
        return hosts

    def list_all_routers_on_hosting_devices(self, context):
        query = context.session.query(
            l3_models.RouterHostingDeviceBinding.router_id)
        query = query.filter(
            l3_models.RouterHostingDeviceBinding.hosting_device_id !=
            expr.null())
        router_ids = [item[0] for item in query]
        if router_ids:
            return self.get_sync_data_ext(context, router_ids=router_ids,
                                          active=True)
        else:
            return []

    def get_number_of_agents_for_scheduling(self, context):
        """Return number of agents on which the router will be scheduled."""

        num_agents = len(self.get_l3_agents(context, active=True,
            filters={'agent_modes': [bc.constants.L3_AGENT_MODE_LEGACY,
                                     bc.constants.L3_AGENT_MODE_DVR_SNAT]}))
        max_agents = cfg.CONF.max_l3_agents_per_router
        if max_agents:
            if max_agents > num_agents:
                LOG.info("Number of active agents lower than "
                         "max_l3_agents_per_router. L3 agents "
                         "available: %s", num_agents)
            else:
                num_agents = max_agents

        return num_agents



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from itertools import groupby
import netaddr
from operator import itemgetter
import time
import uuid

from oslo_config import cfg
from oslo_log import log as logging

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cpnr import cpnr_client
from networking_cisco.plugins.cisco.cpnr import dhcpopts
from neutron.agent.linux import dhcp

LOG = logging.getLogger(__name__)
RELOAD_TIMEOUT = 120


class Network(object):

    def __init__(self):
        self.vpn = Vpn()
        self.view = View()
        self.forward_zones = {}
        self.scopes = {}
        self.client_entries = {}
        self.reverse_zones = {}
        self.hosts = {}

    @classmethod
    def from_neutron(cls, network):
        net = Network()
        net.vpn = Vpn.from_neutron(network)
        net.view = View.from_neutron(network)
        fz = ForwardZone.from_neutron(network)
        net.forward_zones[fz.data['origin']] = fz
        # Ignore DHCP-disabled and IPv6 subnets
        ipv4_subnet_ids = cls.filter_ipv4_subnets(network.subnets)
        for subnet in network.subnets:
            if subnet.id not in ipv4_subnet_ids:
                continue
            net.scopes[subnet.id] = Scope.from_neutron(network, subnet)
            net.reverse_zones[subnet.id] = ReverseZone.from_neutron(network,
                                                                    subnet)
        for port in network.ports:
            ipv4_port = False
            for ip in port.fixed_ips:
                if ip.subnet_id in ipv4_subnet_ids:
                    ipv4_port = True
                    # Update hosts with only IPv4 fixed IPs
                    hostname = Host.addr_to_hostname(ip.ip_address)
                    net.hosts[hostname] = Host.from_neutron(network,
                                                            ip.ip_address)
            if ipv4_port:
                net.client_entries[port.id] = ClientEntry.from_neutron(
                    network, port)
        return net

    def create(self):
        self.vpn.create()
        self.view.create()
        for fz in self.forward_zones.values():
            fz.create()
        for scope in self.scopes.values():
            scope.create()
        for ce in self.client_entries.values():
            ce.create()
        for rz in self.reverse_zones.values():
            rz.create()
        for host in self.hosts.values():
            host.create()

    def update(self, new):
        self.vpn.update(new.vpn)
        self.view.update(new.view)
        self._update_all(self.forward_zones, new.forward_zones)
        self._update_all(self.scopes, new.scopes)
        self._update_all(self.client_entries, new.client_entries)
        self._update_all(self.reverse_zones, new.reverse_zones)
        self._update_all(self.hosts, new.hosts)

    def _update_all(self, old, new):
        deleted_keys = set(old.keys()) - set(new.keys())
        deleted = (old[key] for key in deleted_keys)
        for d in deleted:
            d.delete()
        created_keys = set(new.keys()) - set(old.keys())
        created = (new[key] for key in created_keys)
        for c in created:
            c.create()
        updated_keys = set(old.keys()) & set(new.keys())
        updated = ((old[key], new[key]) for key in updated_keys)
        for o, n in updated:
            o.update(n)

    def delete(self):
        for host in self.hosts.values():
            host.delete()
        for rz in self.reverse_zones.values():
            rz.delete()
        for ce in self.client_entries.values():
            ce.delete()
        for scope in self.scopes.values():
            scope.delete()
        for fz in self.forward_zones.values():
            fz.delete()
        self.view.delete()
        self.vpn.delete()

    @staticmethod
    def filter_ipv4_subnets(subnets):
        return [subnet.id for subnet in subnets
                if subnet.enable_dhcp and subnet.ip_version == 4]


class Vpn(object):

    def __init__(self, data=None):
        self.data = data or {}

    @classmethod
    def from_neutron(cls, network):
        vpnid = cls.net_to_vpn_id(network.id)
        vpnrfc = cls.net_to_vpn_rfc(network.id)
        data = {'name': network.id,
                'description': network.id,
                'id': vpnid,
                'vpnId': vpnrfc}
        return cls(data)

    @classmethod
    def from_pnr(cls, vpn):
        attrs = ['name', 'description', 'id', 'vpnId']
        vpn = {k: vpn[k] for k in attrs if k in vpn}
        return cls(vpn)

    @staticmethod
    def net_to_vpn_id(netid):
        netid = uuid.UUID(netid)
        return str(netid.int & 0x7fffffff)

    @staticmethod
    def net_to_vpn_rfc(netid):
        netid = uuid.UUID(netid)
        return (netid.hex[-14:-8].lstrip('0') + ':' +
            netid.hex[-8:].lstrip('0'))

    def create(self):
        pnr = _get_client()
        pnr.update_vpn(self.data['name'], self.data)

    def update(self, new):
        pnr = _get_client()
        if self.data != new.data:
            pnr.update_vpn(new.data['name'], new.data)
            self.data = new.data

    def delete(self):
        pnr = _get_client()
        pnr.delete_vpn(self.data['name'])


class Scope(object):

    def __init__(self, data=None):
        self.data = data or {}

    @classmethod
    def from_neutron(cls, network, subnet):
        vpnid = Vpn.net_to_vpn_id(network.id)
        range_list = cls._from_port_list_to_range(network.ports, subnet.id)
        policy = Policy.from_neutron_subnet(network, subnet)
        ipv4_subnets = [sub for sub in network.subnets
                        if sub.enable_dhcp and sub.ip_version == 4]
        primary_subnet = ipv4_subnets[0].cidr
        if (subnet.cidr == primary_subnet):
            data = {'name': subnet.id,
                    'vpnId': vpnid,
                    'subnet': subnet.cidr,
                    'rangeList': range_list,
                    'restrictToReservations': 'enabled',
                    'embeddedPolicy': policy.data}
        else:
            data = {'name': subnet.id,
                    'vpnId': vpnid,
                    'subnet': subnet.cidr,
                    'rangeList': range_list,
                    'restrictToReservations': 'enabled',
                    'embeddedPolicy': policy.data,
                    'primarySubnet': primary_subnet}
        return cls(data)

    @classmethod
    def from_pnr(cls, scope):
        attrs = ['name', 'vpnId', 'subnet', 'rangeList',
                 'restrictToReservations', 'embeddedPolicy',
                 'primarySubnet']
        scope = {k: scope[k] for k in attrs if k in scope}
        return cls(scope)

    @staticmethod
    def _from_port_list_to_range(ports, subnet_id):
        # List of IP addresses as integers
        ip_addrs_int = []
        for port in ports:
            for ip in port.fixed_ips:
                if ip.subnet_id != subnet_id:
                    continue
                ip_addrs_int.append(int(netaddr.IPAddress(ip.ip_address)))
        # Group all consecutive addresses and create ranges
        ip_ranges = []
        for key, group in groupby(enumerate(sorted(ip_addrs_int)),
                                  lambda item: item[0] - item[1]):
            ip_range = list(map(itemgetter(1), group))
            min_range = str(netaddr.IPAddress(ip_range[0]))
            max_range = str(netaddr.IPAddress(ip_range[-1]))
            ip_ranges.append((min_range, max_range))
        range_item = [{'start': start, 'end': end}
                      for (start, end) in ip_ranges]
        range_list = {'RangeItem': range_item}
        return range_list

    def create(self):
        pnr = _get_client()
        pnr.update_scope(self.data['name'], self.data)

    def update(self, new):
        pnr = _get_client()
        if self.data != new.data:
            pnr.update_scope(new.data['name'], new.data)
            self.data = new.data

    def delete(self):
        pnr = _get_client()
        pnr.delete_scope(self.data['name'])


class ClientEntry(object):

    def __init__(self, data=None, vpnid='0'):
        self.data = data or {}
        self.vpnid = vpnid

    @classmethod
    def from_neutron(cls, network, port):
        vpnid = Vpn.net_to_vpn_id(network.id)
        name = cls._gen_client_entry_name(network.id, port.mac_address)

        # Retain only IPv4 fixed IPs
        ipv4_subnet_ids = Network.filter_ipv4_subnets(network.subnets)
        fixed_ips_v4 = [fixed_ip for fixed_ip in port.fixed_ips
                        if fixed_ip.subnet_id in ipv4_subnet_ids]

        hostname = ('host-%s' %
                   (fixed_ips_v4[0].ip_address.replace('.', '-'),))
        addrs = {'stringItem': [alloc.ip_address for alloc in fixed_ips_v4]}
        policy = Policy.from_neutron_port(network, port)
        netportid = "%s+%s" % (network.id, port.id)
        data = {'clientClassName': 'openstack-client-class',
                'name': name,
                'hostName': hostname,
                'domainName': cfg.CONF.dns_domain,
                'reservedAddresses': addrs,
                'embeddedPolicy': policy.data,
                'userDefined': netportid}
        return cls(data, vpnid)

    @classmethod
    def from_pnr(cls, ce):
        attrs = ['clientClassName', 'name', 'hostName', 'domainName',
                 'reservedAddresses', 'embeddedPolicy', 'userDefined']
        ce = {k: ce[k] for k in attrs if k in ce}
        (netid, _, _) = ce['userDefined'].partition('+')
        vpnid = Vpn.net_to_vpn_id(netid)
        return cls(ce, vpnid)

    @staticmethod
    def _gen_client_entry_name(netid, macaddr):
        netid = uuid.UUID(netid)
        vpnid = netid.hex[-14:-8] + ':' + netid.hex[-8:]
        vpnid = vpnid.replace(':', '')
        vpnid = ':'.join(vpnid[i:i + 2] for i in range(0, len(vpnid), 2))
        return '01:' + vpnid + ":" + macaddr

    def create(self):
        pnr = _get_client()
        pnr.update_client_entry(self.data['name'], self.data)

    def update(self, new):
        pnr = _get_client()
        if self.data != new.data:
            pnr.update_client_entry(new.data['name'], new.data)
            self.data = new.data

    def delete(self):
        pnr = _get_client()
        pnr.delete_client_entry(self.data['name'])
        for addr in self.data['reservedAddresses']['stringItem']:
            pnr.release_address(addr, self.vpnid)


class Policy(object):

    def __init__(self, data=None):
        self.data = data or {}

    @classmethod
    def from_neutron_subnet(cls, network, subnet):
        options = []
        if subnet.gateway_ip:
            options.append(('routers', subnet.gateway_ip))
        dns_servers = cls._normalize_dns_nameservers(network, subnet)
        if dns_servers:
            options.append(('domain-name-servers', dns_servers))
        static_routes_str = cls._normalize_host_routes(network, subnet)
        static_routes = ""
        if static_routes_str:
            static_routes = dhcpopts.format_for_options(
                            'classless-static-routes', static_routes_str)
        if static_routes:
            options.append(('classless-static-routes', static_routes))
        extra_options = {'dhcp-lease-time': str(cfg.CONF.dhcp_lease_duration),
                         'domain-name': cfg.CONF.dns_domain}
        for option in extra_options.items():
            options.append(option)

        opt_list = []
        for name, val in options:
            opt = dhcpopts.format_for_pnr(name, val)
            if opt:
                opt_list.append(opt)

        if opt_list:
            data = {'optionList': {'OptionItem': opt_list}}
        else:
            data = {'optionList': {'list': []}}
        return cls(data)

    @classmethod
    def from_neutron_port(cls, network, port):
        opt_list = []
        if hasattr(port, 'extra_dhcp_opts'):
            for opt in port.extra_dhcp_opts:
                opt = dhcpopts.format_for_pnr(opt.opt_name, opt.opt_value)
                if opt:
                    opt_list.append(opt)

        if opt_list:
            data = {'optionList': {'OptionItem': opt_list}}
        else:
            data = {'optionList': {'list': []}}
        return cls(data)

    @classmethod
    def _normalize_dns_nameservers(cls, network, subnet):
        if not subnet.dns_nameservers:
            for ip in cls._iter_dhcp_ips(network, subnet):
                subnet.dns_nameservers.append(ip)
        return ','.join(subnet.dns_nameservers)

    @classmethod
    def _normalize_host_routes(cls, network, subnet):
        host_routes = []
        isolated_subnets = dhcp.Dnsmasq.get_isolated_subnets(network)
        if (isolated_subnets[subnet.id] and
           cfg.CONF.enable_isolated_metadata and
           subnet.ip_version == 4):
            class HostRoute(object):
                pass
            for ip in cls._iter_dhcp_ips(network, subnet):
                hr = HostRoute()
                hr.nexthop = ip
                hr.destination = '%s/32' % (dhcp.METADATA_DEFAULT_IP,)
                host_routes.append(hr)
        host_routes.extend(subnet.host_routes)
        encoded_routes = []
        for hr in host_routes:
            (subnet, _, mask) = hr.destination.partition("/")
            sigbytes = ((int(mask) - 1) / 8) + 1
            prefix = '.'.join(subnet.split('.')[:int(sigbytes)])
            destination = mask + '.' + prefix
            encoded_routes.append(destination + ' ' + hr.nexthop)
        return ','.join(encoded_routes)

    @staticmethod
    def _iter_dhcp_ips(net, subnet):
        for port in net.ports:
            if port.device_owner != bc.constants.DEVICE_OWNER_DHCP:
                continue
            for ip in port.fixed_ips:
                if ip.subnet_id != subnet.id:
                    continue
                yield ip.ip_address


class View(object):

    def __init__(self, data=None):
        self.data = data or {}

    @classmethod
    def from_neutron(cls, network):
        viewid = cls.net_to_view_id(network.id)
        data = {'name': network.id,
                'viewId': viewid,
                'priority': viewid}
        return cls(data)

    @classmethod
    def from_pnr(cls, view):
        attrs = ['name', 'viewId', 'priority']
        view = {k: view[k] for k in attrs if k in view}
        return cls(view)

    @staticmethod
    def net_to_view_id(netid):
        netid = uuid.UUID(netid)
        return str(netid.int & 0x7fffffff)

    def create(self):
        pnr = _get_client()
        pnr.update_dns_view(self.data['name'], self.data)

    def update(self, new):
        pnr = _get_client()
        if self.data != new.data:
            pnr.update_dns_view(new.data['name'], new.data)
            self.data = new.data

    def delete(self):
        pnr = _get_client()
        pnr.delete_dns_view(self.data['name'])


class ForwardZone(object):

    def __init__(self, data=None):
        self.data = data or {}

    @classmethod
    def from_neutron(cls, network):
        email = cfg.CONF.cisco_pnr.admin_email.replace('@', '.') + '.'
        viewid = View.net_to_view_id(network.id)
        data = {'origin': cfg.CONF.dns_domain + '.',
                'nameservers': {'stringItem': ['localhost.']},
                'ns': 'localhost.',
                'person': email,
                'serial': '1',
                'viewId': viewid}
        return cls(data)

    @classmethod
    def from_pnr(cls, zone):
        attrs = ['origin', 'nameservers', 'ns', 'person', 'serial', 'viewId']
        zone = {k: zone[k] for k in attrs if k in zone}
        return cls(zone)

    def create(self, retry_count=10):
        pnr = _get_client()
        # When many CCMZone objects are created at the same time,
        # HTTP error code 500 is returned (AX_CCM_DB_UNAVAIL) sometimes.
        # Hence, need to retry updating failed CCMZone object.
        for i in range(retry_count):
            try:
                pnr.update_ccm_zone(self.data['origin'], self.data,
                                    viewid=self.data['viewId'])
                break
            except Exception:
                time.sleep(1)

    def update(self, new):
        pnr = _get_client()
        if self.data != new.data:
            pnr.update_ccm_zone(new.data['origin'], new.data,
                                viewid=self.data['viewId'])
            self.data = new.data

    def delete(self):
        pnr = _get_client()
        pnr.delete_ccm_zone(self.data['origin'], viewid=self.data['viewId'])


class ReverseZone(object):

    def __init__(self, data=None):
        self.data = data or {}

    @classmethod
    def from_neutron(cls, network, subnet):
        octets = subnet.cidr.partition('/')[0].split('.')
        rzone = '%s.%s.%s.in-addr.arpa.' % (octets[2], octets[1], octets[0])
        email = cfg.CONF.cisco_pnr.admin_email.replace('@', '.') + '.'
        viewid = View.net_to_view_id(network.id)
        data = {'origin': rzone,
                'nameservers': {'stringItem': ['localhost.']},
                'ns': 'localhost.',
                'person': email,
                'serial': '1',
                'viewId': viewid,
                'description': subnet.id}
        return cls(data)

    @classmethod
    def from_pnr(cls, rzone):
        attrs = ['origin', 'nameservers', 'ns', 'person',
                 'serial', 'viewId', 'description']
        rzone = {k: rzone[k] for k in attrs if k in rzone}
        return cls(rzone)

    def create(self):
        pnr = _get_client()
        pnr.update_ccm_reverse_zone(self.data['origin'], self.data,
                                    viewid=self.data['viewId'])

    def update(self, new):
        pnr = _get_client()
        if self.data != new.data:
            pnr.update_ccm_reverse_zone(new.data['origin'], new.data,
                                        viewid=self.data['viewId'])
            self.data = new.data

    def delete(self):
        pnr = _get_client()
        pnr.delete_ccm_reverse_zone(self.data['origin'],
                                    viewid=self.data['viewId'])


class Host(object):

    def __init__(self, data=None, viewid=0):
        self.data = data or {}
        self.viewid = viewid

    @classmethod
    def from_neutron(cls, network, addr):
        viewid = View.net_to_view_id(network.id)
        data = {'name': cls.addr_to_hostname(addr),
                'zoneOrigin': cfg.CONF.dns_domain + '.',
                'addrs': {'stringItem': [addr]}}
        return cls(data, viewid)

    @classmethod
    def from_pnr(cls, host, viewid):
        attrs = ['name', 'zoneOrigin', 'addrs']
        host = {k: host[k] for k in attrs if k in host}
        return cls(host, viewid)

    @staticmethod
    def addr_to_hostname(addr):
        return 'host-%s' % addr.replace('.', '-')

    def create(self):
        pnr = _get_client()
        pnr.update_ccm_host(self.data['name'], self.data,
                            viewid=self.viewid,
                            zoneid=self.data['zoneOrigin'])

    def update(self, new):
        pnr = _get_client()
        if self.data != new.data:
            pnr.update_ccm_host(new.data['name'], new.data,
                                viewid=self.viewid,
                                zoneid=self.data['zoneOrigin'])
            self.data = new.data

    def delete(self):
        pnr = _get_client()
        pnr.delete_ccm_host(self.data['name'], viewid=self.viewid,
                            zoneid=self.data['zoneOrigin'])


def _get_client():
    global _pnr_client
    if '_pnr_client' not in globals():
        (addr, _, port) = cfg.CONF.cisco_pnr.http_server.rpartition(":")
        (scheme, _, addr) = addr.rpartition('://')
        if not scheme:
            scheme = 'http'
        user = cfg.CONF.cisco_pnr.http_username
        pswd = cfg.CONF.cisco_pnr.http_password
        insecure = cfg.CONF.cisco_pnr.insecure
        _pnr_client = cpnr_client.CpnrClient(scheme, addr, port, user,
                                             pswd, insecure)
    return _pnr_client


def configure_pnr():
    dhcp_log_settings = ','.join(cfg.CONF.cisco_pnr.dhcp_log_settings)
    dhcp_server_config = {'clientClass': 'enabled',
                          'clientClassLookupId': '"openstack-client-class"',
                          'deleteOrphanedLeases': 'true',
                          'name': 'DHCP',
                          'logSettings': dhcp_log_settings}
    client_class = {'clientLookupId':
                    '(concat (request option 82 151) (request chaddr))',
                    'name': 'openstack-client-class'}
    dns_log_settings = ','.join(cfg.CONF.cisco_pnr.dns_log_settings)
    dns_server_config = {'localPortNum': '5353',
                         'hybridMode': 'true',
                         'hybridOverride': 'exceptions-override-zones',
                         'logSettings': dns_log_settings}
    dns_addrs = [{'addr': a} for a in cfg.CONF.cisco_pnr.upstream_dns_servers]
    dns_forwarder = {'addr': {'DnsIPnPortItem': dns_addrs},
                     'name': '.'}
    email = cfg.CONF.cisco_pnr.admin_email.replace('@', '.')
    dns_zone = {'origin': '.',
                'nameservers': {'stringItem': ['localhost.']},
                'ns': 'localhost.',
                'person': email,
                'serial': '1'}
    try:
        pnr = _get_client()
        pnr.update_dhcp_server(dhcp_server_config)
        pnr.update_client_class(client_class['name'], client_class)
        pnr.update_dns_server(dns_server_config)
        pnr.update_dns_forwarder('%%32%45', dns_forwarder)
        pnr.update_ccm_zone('%%32%45', dns_zone)
    except cpnr_client.CpnrException:
        LOG.error("Failed to configure CPNR DHCP Server and Client Class")


def recover_networks():
    networks = {}
    try:
        networks = _unsafe_recover_networks()
    except Exception:
        LOG.exception("Failed to recover networks. "
                      "CPNR may be unreachable")
    return networks


def _unsafe_recover_networks():
    pnr = _get_client()
    networks = {}
    try:
        for vpn in pnr.get_vpns():
            net = Network()
            netid = vpn['name']
            net.vpn = Vpn.from_pnr(vpn)
            networks[netid] = net
            try:
                for scope in pnr.get_scopes(vpn['id']):
                    net.scopes[scope['name']] = Scope.from_pnr(scope)
            except Exception:
                LOG.exception('Failed to read back scopes for '
                              'network %s', netid)
        for ce in pnr.get_client_entries():
            (netid, _, portid) = ce['userDefined'].partition('+')
            if netid not in networks:
                continue
            net = networks[netid]
            net.client_entries[portid] = ClientEntry.from_pnr(ce)
        for view in pnr.get_dns_views():
            netid = view['name']
            if netid not in networks:
                continue
            net = networks[netid]
            net.view = View.from_pnr(view)
            viewid = view['viewId']
            try:
                for fzone in pnr.get_ccm_zones(viewid=viewid):
                    domain = fzone['origin']
                    net.forward_zones[domain] = ForwardZone.from_pnr(fzone)
                for rz in pnr.get_ccm_reverse_zones(viewid=viewid):
                    subid = rz['description']
                    net.reverse_zones[subid] = ReverseZone.from_pnr(rz)
                for host in pnr.get_ccm_hosts(viewid=viewid, zoneid=domain):
                    net.hosts[host['name']] = Host.from_pnr(host, viewid)
            except Exception:
                LOG.exception('Failed to read back PNR data for '
                              'network %(network)s view %(view)s',
                              {'network': netid, 'view': viewid})
    except Exception:
        LOG.exception('Failed to recover networks from PNR')
    return networks


def get_version():
    version = 0
    try:
        pnr = _get_client()
        verstr = pnr.get_version()
        version = verstr.split()[2]
    except cpnr_client.CpnrException:
        LOG.warning("Failed to obtain CPNR version number")
    except Exception:
        LOG.warning("Failed to parse CPNR version number")
    LOG.debug("CPNR version: %s", version)
    return version


def reload_needed():
    pnr = _get_client()
    return pnr.reload_needed()


def reload_server(timeout=RELOAD_TIMEOUT):
    pnr = _get_client()
    pnr.reload_server()
    reloaded_time = time.time()
    while time.time() - reloaded_time < timeout:
        try:
            if (pnr.get_version() != '' and
               'name' in pnr.get_dhcp_server() and
               'name' in pnr.get_dns_server()):
                return
        except Exception:
            time.sleep(1)
            continue
    LOG.warning("PNR timed out after reload, "
                "timeout: %s seconds", timeout)



# Copyright 2015 Cisco Systems.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function

from networking_cisco._i18n import _

from neutronclient.common import extension
from neutronclient.neutron import v2_0 as neutronV20
from neutronclient.neutron.v2_0 import router

from networking_cisco.neutronclient import hostingdevice


R_RESOURCE = 'router'
DEVICE_L3_ROUTERS = '/hosting-device-l3-routers'


class RoutersOnHostingDevice(extension.NeutronClientExtension):
    resource = R_RESOURCE
    resource_plural = '%ss' % resource
    object_path = '/%s' % resource_plural
    resource_path = '/%s/%%s' % resource_plural
    versions = ['2.0']
    allow_names = True


class AddRouterToHostingDevice(extension.ClientExtensionCreate,
                               RoutersOnHostingDevice):
    """Add a router to hosting device."""

    shell_command = 'cisco-hosting-device-router-add'

    def get_parser(self, prog_name):
        parser = super(AddRouterToHostingDevice, self).get_parser(prog_name)
        parser.add_argument(
            'hosting_device',
            help=_('Name or id of the hosting device.'))
        parser.add_argument(
            'router',
            help=_('Name or id of router to add.'))
        return parser

    def execute(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        neutron_client = self.get_client()
        neutron_client.format = parsed_args.request_format
        _id_hd = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'hosting_device', parsed_args.hosting_device)
        _id_r = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'router', parsed_args.router)
        self.add_router_to_hosting_device(neutron_client, _id_hd,
                                          {'router_id': _id_r})
        print(_('Added router \'%(router)s\' to hosting device \'%(hd)s\'') % {
            'router': parsed_args.router, 'hd': parsed_args.hosting_device},
            file=self.app.stdout, end='')
        return [], []

    def add_router_to_hosting_device(self, client, hosting_device_id, body):
        """Adds a router to hosting device."""
        res_path = hostingdevice.HostingDevice.resource_path
        return client.post((res_path + DEVICE_L3_ROUTERS) %
                           hosting_device_id, body=body)


class RemoveRouterFromHostingDevice(extension.ClientExtensionCreate,
                                    RoutersOnHostingDevice):
    """Remove a router from Hosting Device."""

    shell_command = 'cisco-hosting-device-router-remove'

    def get_parser(self, prog_name):
        parser = super(RemoveRouterFromHostingDevice, self).get_parser(
            prog_name)
        parser.add_argument(
            'hosting_device',
            help=_('Name or id of the hosting device.'))
        parser.add_argument(
            'router',
            help=_('Name or id of router to remove.'))
        return parser

    def execute(self, parsed_args):
        self.log.debug('run(%s)' % parsed_args)
        neutron_client = self.get_client()
        neutron_client.format = parsed_args.request_format
        _id_hd = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'hosting_device', parsed_args.hosting_device)
        _id_r = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'router', parsed_args.router)
        self.remove_router_from_hosting_device(neutron_client, _id_hd, _id_r)
        print(_('Removed router \'%(router)s\' from hosting device \'%(hd)s\'')
              % {'router': parsed_args.router,
                 'hd': parsed_args.hosting_device}, file=self.app.stdout,
              end='')
        return [], []

    def remove_router_from_hosting_device(self, client, hosting_device_id,
                                          router_id):
        """Remove a router from hosting_device."""
        res_path = hostingdevice.HostingDevice.resource_path
        return client.delete((res_path + DEVICE_L3_ROUTERS + "/%s") % (
            hosting_device_id, router_id))


class RoutersOnHostingDeviceList(extension.ClientExtensionList,
                                 RoutersOnHostingDevice):

    shell_command = 'cisco-hosting-device-list-hosted-routers'
    _formatters = {'external_gateway_info':
                   router._format_external_gateway_info}
    list_columns = ['id', 'name', 'external_gateway_info']

    def get_parser(self, prog_name):
        parser = super(RoutersOnHostingDeviceList, self).get_parser(prog_name)
        parser.add_argument(
            'hosting_device',
            help=_('Name or id of the hosting device to query.'))
        return parser

    def call_server(self, neutron_client, search_opts, parsed_args):
        _id = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'hosting_device', parsed_args.hosting_device)
        data = self.list_routers_on_hosting_device(neutron_client, _id,
                                                   **search_opts)
        return data

    def list_routers_on_hosting_device(self, client, hosting_device_id,
                                       **_params):
        """Fetches a list of routers hosted on a hosting device."""
        res_path = hostingdevice.HostingDevice.resource_path
        return client.get((res_path + DEVICE_L3_ROUTERS) %
                          hosting_device_id, params=_params)


HD_RESOURCE = 'hosting_device'
L3_ROUTER_DEVICES = '/l3-router-hosting-devices'


class HostingDeviceHostingRouter(extension.NeutronClientExtension):
    resource = HD_RESOURCE
    resource_plural = '%ss' % resource
    object_path = '/%s' % resource_plural
    resource_path = '/%s/%%s' % resource_plural
    versions = ['2.0']
    allow_names = True


class HostingDeviceHostingRouterList(extension.ClientExtensionList,
                                     HostingDeviceHostingRouter):

    shell_command = 'cisco-router-list-hosting-devices'
    list_columns = ['id', 'name', 'status', 'admin_state_up', 'template_id']

    def get_parser(self, prog_name):
        parser = super(HostingDeviceHostingRouterList, self).get_parser(
            prog_name)
        parser.add_argument('router',
                            help=_('Name or id of router to query.'))
        return parser

    def call_server(self, neutron_client, search_opts, parsed_args):
        _id = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'router', parsed_args.router)
        data = self.list_hosting_devices_hosting_routers(neutron_client, _id,
                                                         **search_opts)
        return data

    def list_hosting_devices_hosting_routers(self, client, router_id,
                                             **_params):
        """Fetches a list of hosting devices hosting a router."""
        return client.get((client.router_path + L3_ROUTER_DEVICES) %
                          router_id, params=_params)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import client
from novaclient import exceptions as nova_exc
from novaclient import utils as n_utils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)

LOG = logging.getLogger(__name__)


SERVICE_VM_LIB_OPTS = [
    cfg.StrOpt('templates_path',
               default='/opt/stack/data/neutron/cisco/templates',
               help=_("Path to templates for hosting devices.")),
    cfg.StrOpt('service_vm_config_path',
               default='/opt/stack/data/neutron/cisco/config_drive',
               help=_("Path to config drive files for service VM instances.")),
]

cfg.CONF.register_opts(SERVICE_VM_LIB_OPTS, "general")


class ServiceVMManager(object):

    def __init__(self, is_auth_v3, user=None, passwd=None,
                 l3_admin_tenant=None, auth_url='', keystone_session=None):

        if is_auth_v3:
            self._nclient = client.Client('2', session=keystone_session,
                                          service_type="compute")
        else:
            self._nclient = client.Client('2', user, passwd, l3_admin_tenant,
                                          auth_url, service_type="compute")

    @property
    def _core_plugin(self):
        try:
            return self._plugin
        except AttributeError:
            self._plugin = bc.get_plugin()
            return self._plugin

    def nova_services_up(self):
        """Checks if required Nova services are up and running.

        returns: True if all needed Nova services are up, False otherwise
        """
        required = set(['nova-conductor', 'nova-cert', 'nova-scheduler',
                       'nova-compute'])
        try:
            services = self._nclient.services.list()
        # There are several individual Nova client exceptions but they have
        # no other common base than Exception, hence the long list.
        except Exception as e:
            LOG.error('Failure determining running Nova services: %s', e)
            return False
        return not bool(required.difference(
            [service.binary for service in services
             if service.status == 'enabled' and service.state == 'up']))

    def get_service_vm_status(self, vm_id):
        try:
            status = self._nclient.servers.get(vm_id).status
        # There are several individual Nova client exceptions but they have
        # no other common base than Exception, hence the long list.
        except Exception as e:
            LOG.error('Failed to get status of service VM instance '
                      '%(id)s, due to %(err)s', {'id': vm_id, 'err': e})
            status = c_constants.SVM_ERROR
        return status

    #TODO(remove fake function later)
    def dispatch_service_vm(self, context, instance_name, vm_image,
                            vm_flavor, hosting_device_drv,
                            credentials_info, connectivity_info,
                            ports=None):
        if True:
            return self.dispatch_service_vm_real(
                context, instance_name, vm_image, vm_flavor,
                hosting_device_drv, credentials_info, connectivity_info, ports)
        else:
            return self.dispatch_service_vm_fake(
                context, instance_name, vm_image, vm_flavor,
                hosting_device_drv, credentials_info, connectivity_info, ports)

    def dispatch_service_vm_real(
            self, context, instance_name, vm_image, vm_flavor,
            hosting_device_drv, credentials_info, connectivity_info,
            ports=None):
        mgmt_port = connectivity_info['mgmt_port']
        nics = [{'port-id': mgmt_port['id']}]
        for port in ports or {}:
            nics.append({'port-id': port['id']})

        try:
            image = n_utils.find_resource(
                bc.get_novaclient_images(self._nclient), vm_image)
            flavor = n_utils.find_resource(self._nclient.flavors, vm_flavor)
        except (nova_exc.CommandError, Exception) as e:
            LOG.error('Failure finding needed Nova resource: %s', e)
            return

        try:
            # Assumption for now is that this does not need to be
            # plugin dependent, only hosting device type dependent.
            files = hosting_device_drv.create_config(context, credentials_info,
                                                     connectivity_info)
        except IOError:
            return

        try:
            server = self._nclient.servers.create(
                instance_name, image.id, flavor.id, nics=nics, files=files,
                config_drive=(files != {}))
        # There are several individual Nova client exceptions but they have
        # no other common base than Exception, therefore the long list.
        except Exception as e:
            LOG.error('Failed to create service VM instance: %s', e)
            return
        return {'id': server.id}

    #TODO(remove fake function later)
    def delete_service_vm(self, context, vm_id):
        if True:
            return self.delete_service_vm_real(context, vm_id)
        else:
            return self.delete_service_vm_fake(context, vm_id)

    def delete_service_vm_real(self, context, vm_id):
        try:
            self._nclient.servers.delete(vm_id)
            return True
        # There are several individual Nova client exceptions but they have
        # no other common base than Exception, therefore the long list.
        except Exception as e:
            LOG.error('Failed to delete service VM instance %(id)s, '
                      'due to %(err)s', {'id': vm_id, 'err': e})
            return False

    # TODO(bobmel): Move this to fake_service_vm_lib.py file with
    # FakeServiceVMManager
    def dispatch_service_vm_fake(self, context, instance_name, vm_image,
                                 vm_flavor, hosting_device_drv,
                                 credentials_info, connectivity_info,
                                 ports=None):
        mgmt_port = connectivity_info['mgmt_port']
        try:
            # Assumption for now is that this does not need to be
            # plugin dependent, only hosting device type dependent.
            hosting_device_drv.create_config(context, credentials_info,
                                             connectivity_info)
        except IOError:
            return

        vm_id = uuidutils.generate_uuid()
        if mgmt_port is not None:
            p_dict = {'port': {'device_id': vm_id,
                               'device_owner': 'nova'}}
            self._core_plugin.update_port(context, mgmt_port['id'], p_dict)

        for port in ports or {}:
            p_dict = {'port': {'device_id': vm_id,
                               'device_owner': 'nova'}}
            self._core_plugin.update_port(context, port['id'], p_dict)

        myserver = {'server': {'adminPass': "MVk5HPrazHcG",
                    'id': vm_id,
                    'links': [{'href': "http://openstack.example.com/v2/"
                                       "openstack/servers/" + vm_id,
                               'rel': "self"},
                              {'href': "http://openstack.example.com/"
                                       "openstack/servers/" + vm_id,
                               'rel': "bookmark"}]}}

        return myserver['server']

    def delete_service_vm_fake(self, context, vm_id):
        result = True

        try:
            ports = self._core_plugin.get_ports(context,
                                                filters={'device_id': [vm_id]})
            for port in ports:
                self._core_plugin.delete_port(context, port['id'])
        except Exception as e:
            LOG.error('Failed to delete service VM %(id)s due to %(err)s',
                      {'id': vm_id, 'err': e})
            result = False
        return result

    def interface_attach(self, vm_id, port_id):
        self._nclient.servers.interface_attach(vm_id, port_id=port_id,
                                               net_id=None, fixed_ip=None)
        LOG.debug('Nova interface add succeeded on VM:%(vm)s for port:%(id)s',
                  {'vm': vm_id, 'id': port_id})

    def interface_detach(self, vm_id, port_id):
        self._nclient.servers.interface_detach(vm_id, port_id)
        LOG.debug('Nova interface detach succeeded on VM:%(vm)s for '
                  'port:%(id)s', {'vm': vm_id, 'id': port_id})

    def vm_interface_list(self, vm_id):
        servers = self._nclient.servers.interface_list(vm_id)
        ips = []
        for s in servers:
            ips.append(s.fixed_ips[0]['ip_address'])
        LOG.info('Interfaces connected on VM:%(vm_id)s is %(ips)s',
                 {'ips': ips, 'vm_id': vm_id})



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import pika
import socket
import sys
import time
import traceback

from oslo_serialization import jsonutils

from networking_cisco.apps.saf.common import dfa_logger as logging

LOG = logging.getLogger(__name__)


class DCNMListener(object):

    """This AMQP client class listens to DCNM's AMQP notification and
    interacts with openstack for further tenant and network information.
    It also communicates with CPNR to populate DHCP data.
    """

    def __init__(self, name, ip, user, password, pqueue=None,
                 c_pri=100, d_pri=100):
        """Create a new instance of AMQP client.

        :param dict params: AMQP configuration parameters,
         e.g. AMQP server ip, port, user name, password, name of AMQP
         exchange and queue for DCNM notification.
        """
        # extract from input params
        self._server_ip = ip
        self._user = user
        self._pwd = password
        self._pq = pqueue
        self._create_pri = c_pri
        self._delete_pri = d_pri
        # exchange, queue name for receiving vCD events.
        self._port = 5672
        self._dcnm_exchange_name = 'DCNMExchange'
        self._dcnm_queue_name = socket.gethostname()
        # specify the key of interest.
        key = ('success.cisco.dcnm.event.auto-config.organization.'
               'partition.network.*')
        self._conn = None
        self.consume_channel = None
        try:
            credentials = None
            if self._user:
                credentials = pika.PlainCredentials(self._user, self._pwd)
            # create connection, channel
            self._conn = pika.BlockingConnection(
                pika.ConnectionParameters(
                    host=self._server_ip,
                    port=self._port,
                    credentials=credentials))

            # create channels for consuming
            self.consume_channel = self._conn.channel()

            # declare vCD exchange
            self.consume_channel.exchange_declare(
                exchange=self._dcnm_exchange_name,
                exchange_type='topic',
                durable=True,
                auto_delete=False)

            result = self.consume_channel.queue_declare(
                queue=self._dcnm_queue_name,
                durable=True,
                auto_delete=False)
            self._dcnm_queue_name = result.method.queue
            self.consume_channel.queue_bind(exchange=self._dcnm_exchange_name,
                                            queue=self._dcnm_queue_name,
                                            routing_key=key)
            # for info only
            msg_count = result.method.message_count
            LOG.debug('The exchange %(exch)s queue %(que)s has '
                      'totally %(cnt)d messages. ',
                      {'exch': self._dcnm_exchange_name,
                       'que': self._dcnm_queue_name, 'cnt': msg_count})

            LOG.debug('DCNM Listener initialization done....')
        except Exception as ex:
            LOG.exception('Failed to initialize DCNMListener %s', ex)

    def _cb_dcnm_msg(self, method, body):
        """Callback function to process DCNM network creation/update/deletion
        message received by AMQP.

        It also communicates with DCNM to extract info for CPNR record
        insertion/deletion.

        :param pika.channel.Channel ch: The channel instance.
        :param pika.Spec.Basic.Deliver method: The basic deliver method
                                               which includes routing key.
        :param pika.Spec.BasicProperties properties: properties
        :param str body: The message body.
        """
        LOG.debug('Routing_key: %(key)s, body: %(body)s.',
                  {'key': method.routing_key, 'body': body})
        partition_keyword = 'auto-config.organization.partition'
        network_keyword = partition_keyword + '.network'

        network_create_key = network_keyword + '.create'
        network_update_key = network_keyword + '.update'
        msg = jsonutils.loads(body)
        LOG.debug('_cb_dcnm_msg: RX message: %s', msg)

        if not msg:
            LOG.debug("error, return")
            return

        url = msg['link']
        url_fields = url.split('/')
        pre_project_name = url_fields[4]
        pre_partition_name = url_fields[6]
        pre_seg_id = url_fields[9]
        data = {"project_name": pre_project_name,
                "partition_name": pre_partition_name,
                "segmentation_id": pre_seg_id}
        if network_create_key in method.routing_key or (
                network_update_key in method.routing_key):
            pri = self._create_pri
            event_type = 'dcnm.network.create'
        else:
            pri = self._delete_pri
            event_type = 'dcnm.network.delete'

        if self._pq is not None:
            payload = (event_type, data)
            self._pq.put((pri, time.ctime, payload))

    def process_amqp_msgs(self):
        """Process AMQP queue messages.

        It connects to AMQP server and calls callbacks to process DCNM events,
        i.e. routing key containing '.cisco.dcnm.', once they arrive in the
        queue.
        """

        LOG.info('Starting process_amqp_msgs...')
        while True:
            (mtd_fr, hdr_fr, body) = (None, None, None)
            try:
                if self.consume_channel:
                    (mtd_fr, hdr_fr, body) = self.consume_channel.basic_get(
                        self._dcnm_queue_name)
                if mtd_fr:
                    # Queue has messages.
                    LOG.info('RX message: %s', body)
                    self._cb_dcnm_msg(mtd_fr, body)
                    self.consume_channel.basic_ack(mtd_fr.delivery_tag)
                else:
                    # Queue is empty.
                    try:
                        self._conn.sleep(1)
                    except AttributeError:
                        time.sleep(1)
            except Exception:
                exc_type, exc_value, exc_tb = sys.exc_info()
                tb_str = traceback.format_exception(exc_type,
                                                    exc_value, exc_tb)
                LOG.exception("Failed to read from queue: %(queue)s "
                              "%(exc_type)s, %(exc_value)s, %(exc_tb)s.", {
                                  'queue': self._dcnm_queue_name,
                                  'exc_type': exc_type,
                                  'exc_value': exc_value,
                                  'exc_tb': tb_str})



# Copyright (c) 2013-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base class driver for SSH and RESTAPI Client.  Some APIs are
called by upper layer code but only apply to either RESTAPIs drivers
or SSH drivers.  Having base class which simply performs 'pass'
is necessary so these two drivers get required actions without
breaking the other driver. Drivers defined in this base are
mostly those called externally.
"""

import os
import threading
import time

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class CiscoNexusBaseDriver(object):
    """Nexus Driver Base Class."""
    def __init__(self, nexus_switches):
        self.nexus_switches = nexus_switches
        self.time_stats = {}

    def keep_ssh_caching(self):
        pass

    def init_ssh_caching(self):
        pass

    def capture_and_print_timeshot(self, start_time, which,
                                   other=99, switch="0.0.0.0"):
        """Determine delta, keep track, and print results."""

        curr_timeout = time.time() - start_time
        if which in self.time_stats:
            self.time_stats[which]["total_time"] += curr_timeout
            self.time_stats[which]["total_count"] += 1
            if (curr_timeout < self.time_stats[which]["min"]):
                self.time_stats[which]["min"] = curr_timeout
            if (curr_timeout > self.time_stats[which]["max"]):
                self.time_stats[which]["max"] = curr_timeout
        else:
            self.time_stats[which] = {
                "total_time": curr_timeout,
                "total_count": 1,
                "min": curr_timeout,
                "max": curr_timeout}
        LOG.debug("NEXUS_TIME_STATS %(switch)s, pid %(pid)d, tid %(tid)d: "
                  "%(which)s_timeout %(curr)f count %(count)d "
                  "average %(ave)f other %(other)d min %(min)f max %(max)f",
                  {'switch': switch,
                  'pid': os.getpid(),
                  'tid': threading.current_thread().ident,
                  'which': which,
                  'curr': curr_timeout,
                  'count': self.time_stats[which]["total_count"],
                  'ave': (self.time_stats[which]["total_time"] /
                          self.time_stats[which]["total_count"]),
                  'other': other,
                  'min': self.time_stats[which]["min"],
                  'max': self.time_stats[which]["max"]})

    def close_session(self, nexus_host):
        pass

    def get_interface_switch(self, nexus_host,
                             intf_type, interface):
        """Get the interface data from host.

        :param nexus_host: IP address of Nexus switch
        :param intf_type:  String which specifies interface type.
                           example: ethernet
        :param interface:  String indicating which interface.
                           example: 1/19
        :returns: response
        """
        return None

    def initialize_all_switch_interfaces(self, interfaces,
                                         switch_ip=None, replay=True):
        """Configure Nexus interface and get port channel number.

        :param interfaces:  List of interfaces for a given switch.
                            ch_grp can be altered as last arg
                            to each interface. If no ch_grp,
                            this arg will be zero.
        :param switch_ip: IP address of Nexus switch
        :param replay: Whether in replay path
        """
        pass

    def get_nexus_type(self, nexus_host):
        """Given the nexus host, get the type of Nexus switch.

        :param nexus_host: IP address of Nexus switch
        :returns: Nexus type
        """
        return None

    def set_all_vlan_states(self, nexus_host, vlanid_range):
        """Set the VLAN states to active."""
        pass

    def start_create_vlan(self):
        """Returns an XML snippet for start of create VLAN."""
        return None, ''

    def get_create_vlan(self, nexus_host, vlanid, vni, conf_str):
        """Returns an XML snippet for create VLAN."""
        return conf_str

    def end_create_vlan(self, conf_str):
        """Returns an XML snippet for terminate of create VLAN."""
        return conf_str

    def create_vlan(self, nexus_host, vlanid, vni):
        """Create a VLAN on a Nexus Switch.

        Creates a VLAN given the VLAN ID and possible VxLAN ID.
        """
        pass

    def delete_vlan(self, nexus_host, vlanid):
        """Delete a VLAN on Nexus Switch given the VLAN ID."""
        pass

    def disable_vlan_on_trunk_int(self, nexus_host, vlanid, intf_type,
                                  interface, is_native):
        """Disable a VLAN on a trunk interface."""
        pass

    def send_edit_string(self, nexus_host, path, confstr,
                         check_to_close_session=True):
        """Sends any XML snippet to Nexus switch."""
        pass

    def send_enable_vlan_on_trunk_int(self, nexus_host, vlanid, intf_type,
                                      interface, is_native):
        """Gathers and sends an interface trunk XML snippet."""
        pass

    def create_and_trunk_vlan(self, nexus_host, vlan_id, intf_type,
                              nexus_port, vni, is_native):
        """Create VLAN and trunk it on the specified ports."""
        pass

    def enable_vxlan_feature(self, nexus_host, nve_int_num, src_intf):
        """Enable VXLAN on the switch."""
        pass

    def disable_vxlan_feature(self, nexus_host):
        """Disable VXLAN on the switch."""
        pass

    def create_nve_member(self, nexus_host, nve_int_num, vni, mcast_group):
        """Add a member configuration to the NVE interface."""
        pass

    def delete_nve_member(self, nexus_host, nve_int_num, vni):
        """Delete a member configuration on the NVE interface."""
        pass



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Based on Neutron L3 Agent scheduler.

import random

from oslo_log import log as logging
from sqlalchemy.orm import exc

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.db.scheduler import cfg_agentschedulers_db

LOG = logging.getLogger(__name__)


class HostingDeviceCfgAgentScheduler(object):
    """A scheduler for Cisco (hosting) device manager service plugin.

    It schedules hosting devices to Cisco cfg agents. The scheduling is a
    simple random selection among qualified candidates.
    """

    def auto_schedule_hosting_devices(self, plugin, context, agent_host):
        """Schedules unassociated hosting devices to Cisco cfg agent.

        Schedules hosting devices to agent running on <agent_host>.
        """
        query = context.session.query(bc.Agent)
        query = query.filter_by(agent_type=c_constants.AGENT_TYPE_CFG,
                                host=agent_host, admin_state_up=True)
        try:
            cfg_agent_db = query.one()
        except (exc.MultipleResultsFound, exc.NoResultFound):
            LOG.debug('No enabled Cisco cfg agent on host %s', agent_host)
            return
        if cfg_agentschedulers_db.CfgAgentSchedulerDbMixin.is_agent_down(
                cfg_agent_db.heartbeat_timestamp):
            LOG.warning('Cisco cfg agent %s is not alive',
                        cfg_agent_db.id)
        return cfg_agent_db

    def schedule_hosting_device(self, plugin, context, hosting_device):
        """Selects Cisco cfg agent that will configure <hosting_device>."""
        active_cfg_agents = plugin.get_cfg_agents(context, active=True)
        if not active_cfg_agents:
            LOG.warning('There are no active Cisco cfg agents')
            # No worries, once a Cisco cfg agent is started and
            # announces itself any "dangling" hosting devices
            # will be scheduled to it.
            return
        LOG.debug('Randomly selecting a Cisco cfg agent among %d candidates' %
                  len(active_cfg_agents))
        return random.choice(active_cfg_agents)


class StingyHostingDeviceCfgAgentScheduler(HostingDeviceCfgAgentScheduler):
    """
    This scheduler will assign a hosting device to an agent that has
    the fewest hosting devices associated to it
    """
    def schedule_hosting_device(self, plugin, context, hosting_device):
        active_cfg_agents = plugin.get_cfg_agents(context, active=True)
        if not active_cfg_agents:
            LOG.warning('There are no active Cisco cfg agents')
            return
        else:
            cfg_agent_hosting_devices = []
            for active_cfg_agent in active_cfg_agents:
                hosting_devices = (
                    plugin.list_hosting_devices_handled_by_cfg_agent(
                        context, active_cfg_agent.id))
                LOG.debug("Cfg agent %s now handling %d hosting devices" % (
                    active_cfg_agent.id,
                    len(hosting_devices.get('hosting_devices'))))
                num_hosting_devices = len(hosting_devices.get(
                    'hosting_devices'))
                cfg_agent_hosting_devices.append((active_cfg_agent,
                                                  num_hosting_devices))
            cfg_agent_hosting_devices = sorted(
                cfg_agent_hosting_devices,
                key=lambda agent_hosting_device: agent_hosting_device[1])
            return cfg_agent_hosting_devices.pop(0)[0]



# Copyright 2014 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Service Constants

import networking_cisco.apps.saf.server.services.constants as services_const

AUTO_NWK_CREATE = True
DEVICE = ''
SCHED_POLICY = 'max_sched'
VLAN_ID_MIN = services_const.VLAN_ID_MIN
VLAN_ID_MAX = services_const.VLAN_ID_MAX
MOB_DOMAIN_NAME = 'md0'
HOST_PROF = 'serviceNetworkUniversalDynamicRoutingESProfile'
HOST_FWD_MODE = 'proxy-gateway'
PART_PROF = 'vrf-common-universal-external-dynamic-ES'
EXT_PROF = 'externalNetworkUniversalDynamicRoutingESProfile'
EXT_FWD_MODE = 'anycast-gateway'
IN_IP_START = '100.100.2.0/24'
IN_IP_END = '100.100.20.0/24'
OUT_IP_START = '200.200.2.0/24'
OUT_IP_END = '200.200.20.0/24'
DUMMY_IP_SUBNET = '9.9.9.0/24'
SERVICE_NAME_EXTRA_LEN = 8

IN_SERVICE_SUBNET = 'FwServiceInSub'
IN_SERVICE_NWK = 'FwServiceInNwk'
SERV_PART_NAME = 'CTX-ext'
OUT_SERVICE_SUBNET = 'FwServiceOutSub'
OUT_SERVICE_NWK = 'FwServiceOutNwk'
DUMMY_SERVICE_RTR = 'DUMMY_SRVC_RTR'
DUMMY_SERVICE_NWK = 'DUMMY_SRVC_NWK'
TENANT_EDGE_RTR = 'Cisco_TenantEdge'
FW_TENANT_EDGE = 'TE'

FW_CR_OP = 'CREATE'
FW_DEL_OP = 'DELETE'

RESULT_FW_CREATE_INIT = 'FAB_CREATE_PEND'
RESULT_FW_CREATE_DONE = 'FAB_CREATE_DONE'
RESULT_FW_DELETE_INIT = 'FAB_DELETE_PEND'
RESULT_FW_DELETE_DONE = 'FAB_DELETE_DONE'

FW_CONST = 'Firewall'
INIT_STATE_STR = 'INIT'

OS_IN_NETWORK_CREATE_FAIL = 'OS_IN_NETWORK_CREATE_FAIL'
OS_INIT_STATE = OS_IN_NETWORK_CREATE_FAIL
OS_IN_NETWORK_CREATE_SUCCESS = 'OS_IN_NETWORK_CREATE_SUCCESS'
OS_OUT_NETWORK_CREATE_FAIL = 'OS_OUT_NETWORK_CREATE_FAIL'
OS_OUT_NETWORK_CREATE_SUCCESS = 'OS_OUT_NETWORK_CREATE_SUCCESS'
OS_DUMMY_RTR_CREATE_FAIL = 'OS_DUMMY_RTR_CREATE_FAIL'
OS_DUMMY_RTR_CREATE_SUCCESS = 'OS_DUMMY_RTR_CREATE_SUCCESS'
OS_CREATE_SUCCESS = OS_DUMMY_RTR_CREATE_SUCCESS
DCNM_IN_NETWORK_CREATE_FAIL = 'DCNM_IN_NETWORK_CREATE_FAIL'
DCNM_INIT_STATE = DCNM_IN_NETWORK_CREATE_FAIL
DCNM_IN_NETWORK_CREATE_SUCCESS = 'DCNM_IN_NETWORK_CREATE_SUCCESS'
DCNM_IN_PART_UPDATE_FAIL = 'DCNM_IN_PART_UPDATE_FAIL'
DCNM_IN_PART_UPDATE_SUCCESS = 'DCNM_IN_PART_UPDATE_SUCCESS'
DCNM_OUT_PART_CREATE_FAIL = 'DCNM_OUT_PART_CREATE_FAIL'
DCNM_OUT_PART_CREATE_SUCCESS = 'DCNM_OUT_PART_CREATE_SUCCESS'
DCNM_OUT_NETWORK_CREATE_FAIL = 'DCNM_OUT_NETWORK_CREATE_FAIL'
DCNM_OUT_NETWORK_CREATE_SUCCESS = 'DCNM_OUT_NETWORK_CREATE_SUCCESS'
DCNM_OUT_PART_UPDATE_FAIL = 'DCNM_OUT_PART_UPDATE_FAIL'
DCNM_OUT_PART_UPDATE_SUCCESS = 'DCNM_OUT_PART_UPDATE_SUCCESS'
DCNM_CREATE_SUCCESS = DCNM_OUT_PART_UPDATE_SUCCESS
# FABRIC_PREPARE_SUCCESS = DCNM_OUT_PART_UPDATE_SUCCESS
FABRIC_PREPARE_SUCCESS = 'FABRIC_PREPARE_SUCCESS'

OS_IN_NETWORK_DEL_FAIL = 'OS_IN_NETWORK_DEL_FAIL'
OS_IN_NETWORK_DEL_SUCCESS = 'OS_IN_NETWORK_DEL_SUCCESS'
OS_OUT_NETWORK_DEL_FAIL = 'OS_OUT_NETWORK_DEL_FAIL'
OS_OUT_NETWORK_DEL_SUCCESS = 'OS_OUT_NETWORK_DEL_SUCCESS'
OS_DUMMY_RTR_DEL_FAIL = 'OS_DUMMY_RTR_DEL_FAIL'
OS_DUMMY_RTR_DEL_SUCCESS = 'OS_DUMMY_RTR_DEL_SUCCESS'
OS_DEL_SUCCESS = 'OS_DUMMY_RTR_DEL_SUCCESS'
DCNM_IN_NETWORK_DEL_FAIL = 'DCNM_IN_NETWORK_DEL_FAIL'
DCNM_IN_NETWORK_DEL_SUCCESS = 'DCNM_IN_NETWORK_DEL_SUCCESS'
DCNM_IN_PART_UPDDEL_FAIL = 'DCNM_IN_PART_UPDDEL_FAIL'
DCNM_IN_PART_UPDDEL_SUCCESS = 'DCNM_IN_PART_UPDDEL_SUCCESS'
DCNM_OUT_PART_DEL_FAIL = 'DCNM_OUT_PART_DEL_FAIL'
DCNM_OUT_PART_DEL_SUCCESS = 'DCNM_OUT_PART_DEL_SUCCESS'
DCNM_OUT_NETWORK_DEL_FAIL = 'DCNM_OUT_NETWORK_DEL_FAIL'
DCNM_OUT_NETWORK_DEL_SUCCESS = 'DCNM_OUT_NETWORK_DEL_SUCCESS'
DCNM_OUT_PART_UPDDEL_FAIL = 'DCNM_OUT_PART_UPDDEL_FAIL'
DCNM_OUT_PART_UPDDEL_SUCCESS = 'DCNM_OUT_PART_UPDDEL_SUCCESS'
DCNM_DELETE_SUCCESS = DCNM_IN_NETWORK_DEL_SUCCESS
INIT = 0
MAX_STATE = FABRIC_PREPARE_SUCCESS  # 17

INIT_STATE = 100
OS_IN_NETWORK_STATE = INIT_STATE + 1
OS_OUT_NETWORK_STATE = OS_IN_NETWORK_STATE + 1
OS_DUMMY_RTR_STATE = OS_OUT_NETWORK_STATE + 1
OS_COMPL_STATE = OS_DUMMY_RTR_STATE
DCNM_IN_NETWORK_STATE = OS_DUMMY_RTR_STATE + 1
DCNM_IN_PART_UPDATE_STATE = DCNM_IN_NETWORK_STATE + 1
DCNM_OUT_PART_STATE = DCNM_IN_PART_UPDATE_STATE + 1
DCNM_OUT_NETWORK_STATE = DCNM_OUT_PART_STATE + 1
DCNM_OUT_PART_UPDATE_STATE = DCNM_OUT_NETWORK_STATE + 1
FABRIC_PREPARE_DONE_STATE = DCNM_OUT_PART_UPDATE_STATE + 1

# The below is for debug display
fw_state_fn_dict = {}
fw_state_fn_dict[INIT_STATE] = 'INIT_STATE'
fw_state_fn_dict[OS_IN_NETWORK_STATE] = 'OS_IN_NETWORK_CREATE_STATE'
fw_state_fn_dict[OS_OUT_NETWORK_STATE] = 'OS_OUT_NETWORK_CREATE_STATE'
fw_state_fn_dict[OS_DUMMY_RTR_STATE] = 'OS_DUMMY_RTR_CREATE_STATE'
fw_state_fn_dict[DCNM_IN_NETWORK_STATE] = 'DCNM_IN_NETWORK_CREATE_STATE'
fw_state_fn_dict[DCNM_IN_PART_UPDATE_STATE] = 'DCNM_IN_PART_UPDATE_STATE'
fw_state_fn_dict[DCNM_OUT_PART_STATE] = 'DCNM_OUT_PART_CREATE_STATE'
fw_state_fn_dict[DCNM_OUT_NETWORK_STATE] = 'DCNM_OUT_NETWORK_CREATE_STATE'
fw_state_fn_dict[DCNM_OUT_PART_UPDATE_STATE] = 'DCNM_OUT_PART_UPDATE_STATE'
fw_state_fn_dict[FABRIC_PREPARE_DONE_STATE] = 'FABRIC_PREPARE_DONE_STATE'

fw_state_fn_del_dict = {}
fw_state_fn_del_dict[INIT_STATE] = 'INIT_STATE'
fw_state_fn_del_dict[OS_IN_NETWORK_STATE] = 'OS_IN_NETWORK_DELETE_STATE'
fw_state_fn_del_dict[OS_OUT_NETWORK_STATE] = 'OS_OUT_NETWORK_DELETE_STATE'
fw_state_fn_del_dict[OS_DUMMY_RTR_STATE] = 'OS_DUMMY_RTR_DELETE_STATE'
fw_state_fn_del_dict[DCNM_IN_NETWORK_STATE] = 'DCNM_IN_NETWORK_DELETE_STATE'
fw_state_fn_del_dict[DCNM_IN_PART_UPDATE_STATE] = 'DCNM_IN_PART_UPDDEL_STATE'
fw_state_fn_del_dict[DCNM_OUT_PART_STATE] = 'DCNM_OUT_PART_DELETE_STATE'
fw_state_fn_del_dict[DCNM_OUT_NETWORK_STATE] = 'DCNM_OUT_NETWORK_DELETE_STATE'
fw_state_fn_del_dict[DCNM_OUT_PART_UPDATE_STATE] = 'DCNM_OUT_PART_UPDDEL_STATE'
fw_state_fn_del_dict[FABRIC_PREPARE_DONE_STATE] = 'FABRIC_PREPARE_DONE_STATE'
